'use strict'

const util = require('./util')
const Logger = util.getLogger(__dirname, __filename)
const TimedCache = require('./timed-cache')
const packageHasher = require('./package-hasher')

const LIVE_AGENT_TTL = 10 * 1000 // ten seconds in ms
const DEAD_AGENT_TTL = 60 * 1000 // 1 minute in ms
const CHECK_INTERVAL = 10 * 1000 // ten seconds
const LOG_INTERVAL = 30 * 1000 // 30 seconds

exports.create = create

function create (context) {
  packageHasher.initialize(context)
  return new AgentManager(context)
}

// Models a collection of N|Solid processes.
class AgentManager {
  constructor (context) {
    this._context = context
    // Note: TimedCache constructor can throw if the ttl argument is invalid
    this._agents = new TimedCache(LIVE_AGENT_TTL, CHECK_INTERVAL)
    this._deadAgents = new TimedCache(DEAD_AGENT_TTL, CHECK_INTERVAL)
    this._commandCBs = new Map()
    this._nextCommandID = util.sequencer()

    // periodically log AgentManager state
    setInterval(() => {
      Logger.info(`Monitoring ${this.list().length} connected NSolid Agents.`)
    }, LOG_INTERVAL).unref()

    this._agents.on('added', (id, agent) => {
      const app = agent.info ? agent.info.app : '... n/a yet ...'
      Logger.debug(`added agent ${id}: app: ${app}`)

      // Agents that start up quickly and send metrics may not yet be
      // subscribed completely, so delaying this request is necessary
      setTimeout(() => {
        // No actual callback required, response is handled by zmq handler.
        if (this.byId(agent.id)) {
          this._context.commandDispatcher.sendCB({
            command: 'packages',
            id: agent.id
          })
        }
      }, 1000).unref()

      if (agent.info == null) {
        Logger.debug('no info for %s', agent.id)
        // Agents that start up quickly and send metrics may not yet be
        // subscribed completely, so delaying this request is necessary
        setTimeout(() => {
          // No actual callback required, response is handled by zmq handler.
          if (this.byId(agent.id)) {
            this._context.commandDispatcher.sendCB({
              command: 'info',
              id: agent.id
            })
          }
        }, 1000).unref()
      }
    })

    this._agents.on('expired', (id, agent) => {
      Logger.debug(`agent ${id} is now considered "missing"`)
      this._context.webhooksManager.sendAgentOffline(agent)
    })
  }

  // Return the agent with specified id.
  byId (id) {
    return this._agents.get(id)
  }

  deadById (id) {
    return this._deadAgents.get(id)
  }

  byPackages (hash) {
    if (!hash) {
      return []
    }

    return this.list().filter(agent => {
      return agent.packagesHash && agent.packagesHash === hash
    })
  }

  // Add an agent.
  upsert (agent) {
    Logger.debug(`adding agent ${agent.id}`)
    this._agents.upsert(agent.id, agent)
  }

  // Remove an agent.
  remove (agent) {
    Logger.debug(`removing agent ${agent.id}`)
    agent.dispose()
    this._agents.delete(agent.id)
    this._context.webhooksManager.sendAgentOffline(agent)
  }

  markDead (agent) {
    this.remove(agent)
    this._deadAgents.upsert(agent.id, agent)
  }

  // Return an array of all the agents.
  list () {
    const agents = this._agents.entries()
    return Object.keys(agents).map((key) => {
      return agents[key]
    })
  }

  reset () {
    this._agents.purge()
    this._deadAgents.purge()
  }
}
