'use strict'

exports.handleRequest = handleRequest

const fs = require('fs')
const path = require('path')

const util = require('../util')
const Logger = util.getLogger(__dirname, __filename)

let Html

function handleRequest (req, res) {
  const context = req.nsStorage.context

  Logger.debug('handling API doc')

  if (Html == null) Html = buildHtml(context)

  res.set('Content-Type', 'text/html')
  res.send(Html)
}

function buildHtml (context) {
  const schema = context.schemas.http
  const baseHtmlFile = path.join(__dirname, 'api-doc-base.html')
  let html = fs.readFileSync(baseHtmlFile, 'utf8')

  for (let method in schema) {
    if (method === 'type') continue

    for (let uri in schema[method]) {
      const qparmsHtml = getQParmsHtml(method, uri, schema[method][uri])

      html = html.replace(
        `%%[query-string-parameters.${method}.${uri}]%%`,
        qparmsHtml
      )
    }
  }

  return html
}

function getQParmsHtml (method, uri, uriSchema) {
  const html = []

  html.push('<table>')
  if (uriSchema.parms.length === 0) html.push('<tr><td>none</td></tr>')

  for (let parm of getParms(method, uri, uriSchema.parms)) {
    html.push(`<tr><td class="qparm">${parm.name}</td><td>${parm.desc}</td></tr>`)
  }

  html.push('</table>')

  if (uriSchema.urlParam != null) {
    const urlParam = uriSchema.urlParam
    html.push('<p>')
    html.push(`The parameter <span class="qparm">${urlParam}</span> may also be used`)
    html.push(`as a url parameter; eg, <tt><b>api/v1/${uri}/${'${' + urlParam + '}'}</b></tt>`)
    html.push('</p>')
  }

  return html.join('\n')
}

// Get the parameters for a specified parameter group
function getParms (method, uri, parmGroups) {
  const result = []

  for (let parm of parmGroups) {
    let required = false

    if (parm[0] === '!') {
      required = true
      parm = parm.substr(1)
    }

    const reqString = required ? ' <b>(required)</b>' : ''

    switch (parm) {
      case 'id':
        result.push({name: 'id', desc: `agent id${reqString}`})
        break
      case 'app':
        result.push({name: 'app', desc: `application name${reqString}`})
        break
      case 'fields':
        result.push({name: 'field', desc: 'output field (may be specified multiple times)'})
        break
      case 'startEnd':
        result.push({name: 'start', desc: 'bound entries starting at specified time'})
        result.push({name: 'end', desc: 'bound entries ending at specified time'})
        break
      case 'interval':
        result.push({name: 'interval', desc: 'generate entries at the specified interval (default: 1 second)'})
        break
      case 'agg':
        result.push({name: 'agg', desc: 'aggregation function; mean, min, max, median'})
        break
      case 'duration':
        result.push({name: 'duration', desc: 'duration in seconds'})
        break
      case 'key':
        result.push({name: 'key', desc: 'settings key'})
        break
      case 'val':
        result.push({name: 'val', desc: 'settings value'})
        break
      case 'name':
        result.push({name: 'name', desc: `custom command name${reqString}`})
        break
      case 'data':
        result.push({name: 'data', desc: 'custom command value - any JSONable value'})
        break
      case 'agentScope':
        result.push({name: 'id', desc: 'agent id'})
        result.push({name: 'app', desc: 'application name'})
        result.push({name: 'hostname', desc: 'hostname'})
        result.push({name: 'tag', desc: 'application tag (may be specified multiple times)'})
        break
      case 'asset':
        result.push({name: 'asset', desc: 'asset id'})
        break
      case 'type':
        result.push({name: 'type', desc: 'asset type: snapshot, profile, etc'})
        break
      case 'firstName':
        result.push({name: 'firstName', desc: 'firstName'})
        break
      case 'lastName':
        result.push({name: 'lastName', desc: 'lastName'})
        break
      case 'email':
        result.push({name: 'email', desc: 'email'})
        break
      case 'company':
        result.push({name: 'company', desc: 'company'})
        break
      case 'optOut':
        result.push({name: 'optOut', desc: 'optOut'})
        break
      case 'license':
        result.push({name: 'license', desc: 'license'})
        break
      default:
        Logger.warn(`unknown parm ${parm} for ${method} ${uri}`)
    }
  }

  return result
}

if (require.main === module) buildBaseDoc()

function buildBaseDoc () {
  const schema = util.readSchema('http')

  const names = {}
  for (let method in schema) {
    if (method === 'type') continue

    for (let name in schema[method]) {
      if (names[name] == null) names[name] = []
      names[name].push(method)
      names[name].sort()
    }
  }

  console.log('<h1>N|Solid Storage HTTP API</h1>')
  console.log('<h2>HTTP endpoints</h2>')

  const namesSorted = Object.keys(names).sort()
  for (let name of namesSorted) {
    for (let method of names[name]) {
      console.log('')
      console.log('<!-- ================================================== -->')
      console.log(`<h3 class="api">${method} /api/v1/${name}</h3>`)
      console.log('<p>')
      console.log('</p>')
      console.log('')
      console.log('<h4>Query String Parameters</h4>')
      console.log(`%%[query-string-parameters.${method}.${name}]%%`)
      console.log('')
      console.log('<h4>Response</h4>')
      console.log('<p>')
      console.log('</p>')
    }
  }
}
