'use strict'

const lodash = require('lodash')

const util = require('../util')

const qsp = require('./query-string-parser')
const EventStream = require('./event-stream')

exports.getHandler = getHandler

const Logger = util.getLogger(__dirname, __filename)

const AllFields = [
  'uptime', 'rss', 'heapTotal', 'heapUsed', 'activeRequests', 'activeHandles',
  'title', 'user', 'cpu', 'freeMem', 'systemUptime', 'load1m', 'load5m', 'id',
  'load15m', 'cpuSpeed', 'app', 'hostname', 'time', 'percentCpuThreshold',
  'percentHeapThreshold', 'totalHeapSizeExecutable', 'totalPhysicalSize',
  'totalAvailableSize', 'heapSizeLimit', 'loopIdlePercent', 'loopEstimatedLag',
  'loopsPerSecond', 'loopAvgTasks', 'loopTotalCount',
  'blockInputOpCount', 'blockOutputOpCount', 'cpuPercent', 'cpuSystemPercent',
  'cpuUserPercent', 'gcForcedCount', 'gcFullCount', 'gcCount', 'gcCpuPercent',
  'gcDurUs99Ptile', 'gcDurUsMedian', 'pageFaultHardCount', 'ctxSwitchInvoluntaryCount',
  'ipcReceivedCount', 'ipcSentCount', 'gcMajorCount', 'signalCount', 'pageFaultSoftCount',
  'swapCount', 'ctxSwitchVoluntaryCount', 'httpClientCount', 'httpClientMedian',
  'httpClient99Ptile', 'httpServerCount', 'httpServerMedian', 'httpServer99Ptile',
  'dnsCount', 'dnsMedian', 'dns99Ptile', 'httpClientAbortCount', 'httpServerAbortCount'
]

const CalculatedFields = [
  'percentHostMemoryUsedByProcess', 'percentHostMemoryUsed'
]

const AllAndCalculatedFields = AllFields.concat(CalculatedFields)

const StringFields = [
  'id', 'title', 'user'
]

// Cause "new" data to be delayed by this duration before it is queried for
// polling requests.  It should be enough buffer that both agents and Influx
// have time to get the data stored before the window is queried.
const RECENT_DELAY = 1000

exports.allFields = AllFields
exports.stringFields = StringFields

exports.handleRequest = handleRequest

function handleRequest (req, res) {
  const context = req.nsStorage.context
  handlerWithContext(context, req, res)
}

// Returns an HTTP request/response handler for the given context.
function getHandler (context) {
  return handler

  function handler (req, res) {
    handlerWithContext(context, req, res)
  }
}

// handle a request the specified context
function handlerWithContext (context, req, res) {
  const qData = qsp.parse('fields agentScope agg startEnd'.split(' '), req.query)
  Logger.debug('query:', req.query)
  Logger.debug('qData:', qData)

  if (qData.errors.length !== 0) {
    const sep = '\n- '
    const msg = `errors in query string parameters:${sep}${qData.errors.join(sep)}`
    res.status(400).send(msg)
    return
  }

  for (let warning of qData.warnings) {
    Logger.warn(`for ${JSON.stringify(req.query)}: ${warning}`)
  }

  let contInterval
  if (qData.end == null) {
    contInterval = setInterval(readNextMetrics, 1000)
    contInterval.unref()
  }

  // Normalize the query data.
  if (qData.fields == null || qData.fields.length === 0) {
    qData.fields = AllAndCalculatedFields
  } else {
    qData.fields.push('id')
  }

  if (qData.start == null) qData.start = Date.now() - 60 * 1000
  if (qData.end == null) qData.end = Date.now() - RECENT_DELAY

  // if (qData.agg == null) qData.agg = 'raw'

  let diff = lodash.difference(qData.fields, AllAndCalculatedFields)
  for (let field of diff) {
    Logger.warn(`for ${JSON.stringify(req.query)}: ignoring invalid field ${field}`)
  }

  qData.fields = lodash.intersection(qData.fields, AllAndCalculatedFields)

  const query = buildQuery(qData)

  let eventStream = req.nsStorage.eventStream

  Logger.debug(`running query: ${query}`)
  readMetrics(context, query, readMetricsDone)

  // read metrics, returning empty data if influx hasn't started yet
  function readMetrics (context, query, cb) {
    if (context.influxdbReader == null) {
      setImmediate(cb, null, '{}')
    } else {
      context.influxdbReader.readMetrics(query, cb)
    }
  }

  function readMetricsDone (err, data) {
    if (err) {
      Logger.error(err, 'error reading metrics')

      if (eventStream == null) {
        res.status(400).send(err.body)
        return
      }

      eventStream.close()
      return
    }

    if (eventStream == null) {
      eventStream = EventStream.create(context, req, res)
    }

    if (eventStream.isClosed()) return

    const events = eventsFromQuery(context, data)
    for (let event of events) {
      eventStream.writeData(event)
    }

    if (contInterval == null) {
      eventStream.close()
    }
  }

  function readNextMetrics () {
    if (eventStream != null && eventStream.isClosed()) {
      clearInterval(contInterval)
      contInterval = null
      return
    }

    qData.start = qData.end
    qData.end = Date.now() - RECENT_DELAY

    const query = buildQuery(qData)

    Logger.debug(`running query: ${query}`)
    readMetrics(context, query, readMetricsDone)
  }
}

// Build the query from the query string data.
function buildQuery (qData) {
  const queryParts = ['SELECT']
  let qFields = lodash.clone(qData.fields)

  if (qData.agg === 'raw') {
    if (qFields.indexOf('id') === -1) qFields.push('id')
  }

  if (qFields.indexOf('percentHostMemoryUsedByProcess') !== -1) qFields.push('rss')
  if (qFields.indexOf('percentHostMemoryUsed') !== -1) qFields.push('freeMem')

  qFields = lodash.without(qFields, 'percentHostMemoryUsedByProcess', 'percentHostMemoryUsed')

  const queryFields = qFields.map(f => fieldRef(f, qData.agg))

  queryParts.push(queryFields)

  queryParts.push('FROM metrics WHERE')

  const whereParts = []

  whereParts.push(`time >= '${util.dateMS2ISO(qData.start)}'`)
  whereParts.push(`time < '${util.dateMS2ISO(qData.end)}'`)

  if (qData.id) {
    whereParts.push(`id::tag = '${qData.id}'`)
  } else {
    if (qData.app) whereParts.push(`app::tag = '${qData.app}'`)
    if (qData.hostname) whereParts.push(`hostname::tag = '${qData.hostname}'`)
  }

  queryParts.push(whereParts.join(' AND '))

  return queryParts.join(' ')
}

const Agg2Fn = {
  min: 'MIN',
  max: 'MAX',
  mean: 'MEAN',
  median: 'MEDIAN'
}

// Return a field reference given the field name and aggregation

const IndexFields = 'id app hostname tags'.split(' ')
function fieldRef (field, agg) {
  if (agg === 'raw') return `"${field}"`

  let fn = Agg2Fn[agg]
  if (fn == null) {
    Logger.error(`invalid aggregator function ${agg}; using 'mean' instead`)
    fn = 'MEAN'
  }

  if (IndexFields.indexOf(field) === -1) {
    return `"${field}" AS "${field}"`
  } else {
    return `"${field}"::tag AS "${field}"`
  }
}

// Create events from the query results.
function eventsFromQuery (context, data) {
  const agentManager = context.agentManager

  try {
    data = JSON.parse(data)
  } catch (err) {
    Logger.warn(err, 'error parsing JSON data from query', data)
    return []
  }

  if (data.results == null) return []
  if (data.results.length === 0) return []
  if (data.results[0].series == null) return []
  if (data.results[0].series.length === 0) return []

  const series = data.results[0].series[0]
  const events = []

  Logger.debug(`eventsFromQuery(): received ${series.values.length} values`)
  for (let i = 0; i < series.values.length; i++) {
    var event = {}
    for (let j = 0; j < series.columns.length; j++) {
      // only add the value if it isn't null
      if (series.values[i][j] == null) {
        Logger.debug(`eventsFromQuery(): value of field '${series.columns[j]}' was null, not added to result`)
      } else {
        event[series.columns[j]] = series.values[i][j]
      }
    }

    const agent = agentManager.byId(event.id)
    if (agent) {
      event = agent.formatMetricsFromQuery(event)
    }

    Logger.debug(`eventsFromQuery(): adding event ${JSON.stringify(event)}`)
    events.push(event)
  }

  return events
}
