'use strict'

const lodash = require('lodash')

const util = require('../util')

exports.syncPollTillDone = syncPollTillDone
exports.getHandlerForCommandWithAsset = getHandlerForCommandWithAsset
exports.getBroadcastFilter = getBroadcastFilter

const Logger = util.getLogger(__dirname, __filename)

const BroadcastFilterQParms = 'app,hostname,tags'.split(/,/g)

// Run fnPoll every interval ms, until Date.now > end or eventStream is closed.
// When finished, call fnDone.
// First run will be after the specified interval.
// It's assumed the fnPoll function is synchronous, so that results of
// calling it multiple times won't result in out-of-order responses.
function syncPollTillDone (eventStream, interval, endDate, fnPoll, fnDone) {
  setTimeout(
    () => _syncPollTillDone(eventStream, interval, endDate, fnPoll, fnDone),
    interval
  ).unref()
}

function _syncPollTillDone (eventStream, interval, endDate, fnPoll, fnDone) {
  let done = false
  if (eventStream.isClosed()) done = true
  if (endDate && Date.now() > endDate) done = true

  if (done) {
    if (fnDone) fnDone()
    return
  }

  fnPoll()

  setTimeout(
    () => _syncPollTillDone(eventStream, interval, endDate, fnPoll, fnDone),
    interval
  ).unref()
}

// Returns an HTTP request/response handler for the given context and command.
function getHandlerForCommandWithAsset (command, getCommandArgs) {
  return (req, res) =>
    handlerForCommandWithAsset(req, res, command, getCommandArgs)
}

// HTTP request/response handler for the given context and command.
function handlerForCommandWithAsset (req, res, command, getCommandArgs) {
  const context = req.nsStorage.context
  const qData = req.nsStorage.qData
  const agentId = qData.id

  if (!agentId) {
    return sendCommandErrorResponse(context, res, {
      status: 400,
      command: command,
      error: {message: 'query parameter id expected'},
      agentId: agentId
    })
  }

  const commandObject = {
    command: command,
    id: agentId
  }

  if (getCommandArgs != null) {
    commandObject.args = getCommandArgs(req)
  }

  // Send the command.
  Logger.debug('sending agent command:', commandObject)
  const requestId = context.commandDispatcher.send(commandObject)

  // On a command error, send appropriate response.
  context.commandDispatcher.on(`error:${requestId}`, err => {
    sendCommandErrorResponse(context, res, {
      status: 500,
      command: command,
      error: {message: err.message},
      agentId: agentId,
      requestId: requestId
    })
  })

  // On a command success, set up the asset handling.
  context.commandDispatcher.on(`data:${requestId}`, (cmdResponse) => {
    Logger.debug(`command received for ${requestId}`, cmdResponse)
  })

  context.assetManager.once(`add:${requestId}`, (asset) => {
    Logger.debug(`asset added, waiting for complete: ${asset}`)
    asset.once('complete', (err) => onAssetComplete(err, asset))
  })

  // Once the asset has completed writing, send the asset over.
  function onAssetComplete (err, asset) {
    Logger.debug(`asset complete for requestId: ${requestId}`, err)
    if (err) {
      return sendCommandErrorResponse(context, res, {
        status: 500,
        command: command,
        error: {message: err.message},
        agentId: agentId,
        requestId: requestId,
        asset: asset
      })
    }

    sendAsset(asset)
  }

  // Send the asset info as the http response
  function sendAsset (asset) {
    const assetURL = `/api/v1/asset?asset=${asset.bulkId}`
    // const cdFile = asset.getDownloadFileName()

    try {
      res.location(assetURL)
      // res.setHeader('Content-disposition', `attachment; filename=${cdFile}`)
      res.status(201)
    } catch (err) {
      Logger.error('Received additional assets data after completion.', err)
      return
    }

    const event = {
      id: agentId,
      time: Date.now(),
      asset: asset.bulkId,
      assetURL: assetURL
    }

    const agent = context.agentManager.byId(agentId)
    if (agent && agent.info) {
      event.app = agent.info.app
      event.tags = agent.info.tags
      event.hostname = agent.info.hostname
      event.pid = agent.info.pid
      event.title = agent.metrics ? agent.metrics.title : undefined
    }

    res.send(event)

    cleanUpCommandRequest(context, requestId, asset)
  }
}

// Send an error response
function sendCommandErrorResponse (context, res, parms) {
  const status = parms.status || 500
  const command = parms.command || 'unknown command'
  const error = parms.error || 'unknown error'
  const agentId = parms.agentId
  const requestId = parms.requestId
  const asset = parms.asset

  Logger.error(`error sending command ${command}:`, error)

  if (res.__nsAlreadySent) return
  res.__nsAlreadySent = true

  cleanUpCommandRequest(context, requestId, asset)

  const body = {}

  if (agentId) body.id = agentId

  body.time = Date.now()
  body.error = error

  res.status(status).send(body)
}

// Cleanup all the command handlers etc we registered
function cleanUpCommandRequest (context, requestId, asset) {
  if (requestId != null) {
    context.commandDispatcher.removeAllListeners(`end:${requestId}`)
    context.commandDispatcher.removeAllListeners(`data:${requestId}`)
    context.commandDispatcher.removeAllListeners(`error:${requestId}`)
  }

  if (asset != null) {
    asset.removeAllListeners('complete')

    const assetId = asset.getAssetId()
    if (assetId != null) {
      context.assetManager.removeAllListeners(`add:${requestId}`)
    }
  }
}

// Return an agent message filter from agentScope querystring data
function getBroadcastFilter (qData) {
  const result = lodash.pick(qData, BroadcastFilterQParms)

  if (lodash.isEmpty(result)) return null

  return result
}
