'use strict'

const lodash = require('lodash')

const httpUtils = require('./http-utils')

exports.run = run

/*
 * Run a command and write each response to the event stream. If an error
 * occurs, end the stream. If we get an `end` event from the command dispatcher,
 * end the stream.
 *
 * @param {object} context
 * @param {EventStream} eventStream
 * @param {object} params
 * @param {function} params.transformer - A function that gets passed a ZeroMQ
 * message, synchronously or asynchronously transforms the message, and passes
 * it to a node-style callback.
 * @returns {string} requestId
 */
const noopTransformer = (data, cb) => cb(null, data)

function run (context, eventStream, qData, params) {
  const command = params.command || 'unknown command'

  if (!eventStream) {
    throw new Error('Can not run a simple command without event stream.')
  }

  const transformer = params.transformer || noopTransformer
  params = lodash.clone(params)
  delete params.transformer

  if (qData.id) {
    params.id = qData.id
  } else {
    const filter = httpUtils.getBroadcastFilter(qData)
    if (filter != null) {
      params.filter = filter
    }
  }

  const requestId = context.commandDispatcher.send(params)

  context.commandDispatcher.on(`error:${requestId}`, err => {
    eventStream.writeError(null, `command ${command}`, err.message)
    return endCommand(requestId, eventStream, context)
  })

  context.commandDispatcher.on(`data:${requestId}`, data => {
    if (eventStream.isClosed()) {
      return endCommand(requestId, eventStream, context)
    }

    if (data) {
      if (data.body === 'PONG') data.body = {pong: true}
      const responseBody = data.body ? lodash.clone(data.body) : {}
      responseBody.id = data.agentId
      responseBody.time = data.time

      transformer(responseBody, (err, transformedData) => {
        if (err) {
          eventStream.writeError(null, `command ${command}`, err.message)
          return endCommand(requestId, eventStream, context)
        }

        eventStream.writeData(transformedData)
        // All POST commands at this point only expect one response if an id is supplied
        if (qData.id != null && !eventStream.isClosed()) eventStream.close()
      })
    }
  })

  context.commandDispatcher.on(`end:${requestId}`, () => {
    endCommand(requestId, eventStream, context)
  })

  return requestId
}

function endCommand (id, eventStream, context) {
  eventStream.close()
  context.commandDispatcher.removeAllListeners(`end:${id}`)
  context.commandDispatcher.removeAllListeners(`data:${id}`)
  context.commandDispatcher.removeAllListeners(`error:${id}`)
}
