'use strict'

const util = require('../util')

exports.ping = ping
exports.write = write
exports.query = query

const Logger = util.getLogger(__dirname, __filename)

const clientRequest = require('client-request')
const encodeForm = require('form-urlencoded')
const encodeQuery = require('querystring').encode

// Write a line to InfluxDB.
function write (port, db, data, cb) {
  Logger.debug(`writing to ${db}`, data)

  const qs = encodeQuery({
    db: db,
    precision: 'n' // nanoseconds; see: https://docs.influxdata.com/influxdb/v0.13/concepts/api/
  })

  clientRequest({
    uri: `http://localhost:${port}/write?${qs}`,
    method: 'POST',
    body: data
  }, onRequest)

  function onRequest (err, res, body) {
    if (err) {
      Logger.error(err, 'error writing data', data)
      if (cb) {
        cb(err)
      }
      return
    }

    if (res.statusCode === 400) {
      Logger.error('syntax error writing data', data, body.toString())
      if (cb) {
        cb(new Error('syntax error writing data'))
      }
      return
    }

    if (res.statusCode !== 204) {
      Logger.error('unknown error writing data', data, body.toString())
      if (cb) {
        cb(new Error('unknown error writing data'))
      }
      return
    }

    if (cb) {
      cb()
    }
  }
}

// Run a query against InfluxDB.
function query (port, db, command, cb) {
  const timeStarted = Date.now()

  const form = encodeForm({
    db: db,
    epoch: 'ms',
    q: command
  })

  clientRequest({
    uri: `http://localhost:${port}/query`,
    method: 'POST',
    body: form,
    headers: {
      'content-type': 'application/x-www-form-urlencoded'
    }
  }, onRequest)

  function onRequest (err, res, body) {
    const timeElapsed = Date.now() - timeStarted
    const statusCode = res ? res.statusCode : 'n/a'
    const message = `query ${db}: ${command}; ${statusCode}; ${timeElapsed} ms`
    if (err) {
      Logger.error(err, message)
    } else {
      Logger.debug(message)
    }

    if (err) return cb(err)

    switch (res.statusCode) {
      case 200: return cb(null, body)
      case 204: return cb(null, body)
    }

    err = new Error('status ' + res.statusCode + '; command: ' + command)
    err.body = '' + body

    Logger.error(`error from query: ${body}`)
    cb(err)
  }
}

// Ping InfluxDB, return the version on success.
function ping (port, cb) {
  clientRequest({
    uri: `http://localhost:${port}/ping`,
    method: 'GET'
  }, onRequest)

  function onRequest (err, res, body) {
    if (err) return cb(err)

    if (res.statusCode !== 204) {
      const err = new Error(`ping status: ${res.statusCode}, body: ${body}`)
      return cb(err)
    }

    const version = res.headers['x-influxdb-version']
    if (version != null) return cb(null, version)

    err = new Error('ping did not return a X-Influxdb-Version header')
    return cb(err)
  }
}
