'use strict'

const fs = require('fs')
const path = require('path')
const childProcess = require('child_process')

const lodash = require('lodash')
const portFinder = require('portfinder')
const asynclib = require('async')

const api = require('./api')
const util = require('../util')

exports.launch = launch

const Logger = util.getLogger(__dirname, __filename)

// Launch an instance of InfluxDB.
function launch (context, cb) {
  const config = context.config

  config.influxdb = {}
  config.influxdb.metaDir = path.join(config.data.dir, 'influxdb', 'meta')
  config.influxdb.dataDir = path.join(config.data.dir, 'influxdb', 'data')
  config.influxdb.walDir = path.join(config.data.dir, 'influxdb', 'wal')

  let err

  err = util.ensureDirExists(config.influxdb.metaDir)
  if (err) return util.cbNextTick(cb, err)

  err = util.ensureDirExists(config.influxdb.dataDir)
  if (err) return util.cbNextTick(cb, err)

  err = util.ensureDirExists(config.influxdb.walDir)
  if (err) return util.cbNextTick(cb, err)

  const opts = {
    host: 'localhost',
    port: 10001
  }
  portFinder.getPorts(2, opts, (err, ports) =>
    gotPorts(err, config, ports, cb)
  )
}

// Capture the ports needed for InfluxDB.
function gotPorts (err, config, ports, cb) {
  if (err) return cb(err)

  config.influxdb.portCluster = ports[0]
  config.influxdb.portHTTP = ports[1]

  configureInfluxDB(config, cb)
}

// Configure InfluxDB.
function configureInfluxDB (config, cb) {
  const logFile = config.logs.influx

  let logFD
  try {
    logFD = fs.openSync(logFile, 'a+')
  } catch (err) {
    Logger.error(err, `error opening InfluxDB log file: ${logFile}`)
    return cb(err)
  }

  Logger.info(`writing InfluxDB log to ${logFile}`)

  const env = {
    INFLUXDB_HOSTNAME: 'localhost',
    INFLUXDB_REPORTING_DISABLED: 'true',
    // following is apparently needed, but not used
    // See https://github.com/nodesource/nsolid-storage/issues/357
    INFLUXDB_BIND_ADDRESS: `localhost:${config.influxdb.portCluster}`,
    INFLUXDB_META_DIR: config.influxdb.metaDir,
    INFLUXDB_DATA_DIR: config.influxdb.dataDir,
    INFLUXDB_DATA_WAL_DIR: config.influxdb.walDir,
    INFLUXDB_ADMIN_ENABLED: false,
    INFLUXDB_HTTP_BIND_ADDRESS: `0.0.0.0:${config.influxdb.portHTTP}`,
    INFLUXDB_DATA_QUERY_LOG_ENABLED: false,
    INFLUXDB_HTTP_LOG_ENABLED: false
  }

  let proc

  try {
    proc = launchInfluxDB(env, logFD)
  } catch (err) {
    Logger.error(err, 'error launching InfluxDB ...')
    return cb(err)
  }

  const influxDBProcess = new InfluxDBProcess(proc, config)

  Logger.debug('waiting for InfluxDB to become ready ...')
  waitForInfluxDB(config, influxDBProcess, influxDBready)

  function influxDBready (err, version) {
    if (err) {
      Logger.error(`error waiting for InfluxDB to become ready: ${err}`)
      return cb(err)
    }

    function dbReady (err) {
      if (err) {
        Logger.error(err, 'error creating continuous queries')
        return cb(err)
      }
      cb(null, influxDBProcess)
    }
    function rpsCreated (err) {
      if (err) {
        Logger.error(err, 'error setting replication policies')
        return cb(err)
      }
      createCQs(config, dbReady)
    }
    function dbCreated (err, body) {
      if (err) {
        Logger.error(err, 'error creating database nsolid')
        return cb(err)
      }
      createRps(config, rpsCreated)
    }
    createDb(config, dbCreated)
  }
}

function createDb (config, cb) {
  const createCommand = 'CREATE DATABASE nsolid'
  api.query(config.influxdb.portHTTP, 'nsolid', createCommand, cb)
}

var policies = [
  {
    name: 'agentData',
    retentionHours: 1,
    default: true
  },
  {
    name: 'one_minute_daily',
    retentionDays: 1
  },
  {
    name: 'one_hour_monthly',
    retentionDays: 31,
    useSettingsDuration: true
  }
]

function createRps (config, cb) {
  const getRetention = 'SHOW RETENTION POLICIES ON nsolid'
  api.query(config.influxdb.portHTTP, 'nsolid', getRetention, (err, body) => {
    if (err) {
      Logger.error(err, 'error checking existing retention policies')
      return cb(err)
    }

    // The result will look something like this:
    // { "results": [ {
    //   "series": [ {
    //     "columns": [ "name", "duration", "shardGroupDuration", "replicaN", "default" ],
    //     "values": [
    //       [ "default", "0", "168h0m0s", 1, false ],
    //       [ "agentData", "720h0m0s", "24h0m0s", 1, true ]
    //     ]
    //   } ]
    // } ] }

    var currentPolicies
    try {
      const response = JSON.parse(body)
      currentPolicies = response.results[0].series[0].values
    } catch (err) {
      Logger.error(err, 'error checking existing retention policies')
    }

    asynclib.eachSeries(policies, (policy, next) => {
      createOrAlterRP(config, currentPolicies, policy, next)
    }, cb)
  })
}

function createOrAlterRP (config, currentPolicies, spec, cb) {
  let exists
  let currentRetentionHours

  // Only the longest RP will be impacted by the user-configurable
  // NSOLID_CONSOLE_DATA_RETENTION_DAYS
  let requestedRetentionHours = spec.useSettingsDuration
    ? (parseInt(config.data.retentionDays, 10) * 24)
    : spec.retentionDays ? (spec.retentionDays * 24) : spec.retentionHours

  const matching = currentPolicies.filter(pol => { return pol[0] === spec.name })
  if (matching.length) {
    exists = true

    const match = matching[0][1].match(/^(\d+)h.*/)
    if (match) {
      currentRetentionHours = parseInt(match[1], 10)
    }
  }

  if (currentRetentionHours != null) {
    Logger.info(`current retention policy for ${spec.name}: ${currentRetentionHours} hour(s)`)
  }

  if (!exists || (currentRetentionHours !== requestedRetentionHours)) {
    Logger.info(`setting retention policy ${spec.name}: ${requestedRetentionHours} hour(s)`)
    const verb = exists ? 'ALTER' : 'CREATE'
    const retentionCommand = `${verb} RETENTION POLICY ${spec.name} ON nsolid ` +
      `DURATION ${requestedRetentionHours}h REPLICATION 1 SHARD DURATION ` +
      `${requestedRetentionHours}h ${spec.default ? 'DEFAULT' : ''}`
    api.query(config.influxdb.portHTTP, 'nsolid', retentionCommand, cb)
  } else {
    cb()
  }
}

var columnAggTypes = {
  mean: [
    'activeHandles', 'activeRequests', 'cpuPercent', 'cpuSpeed',
    'cpuSystemPercent', 'cpuUserPercent', 'freeMem', 'gcCpuPercent',
    'gcDurUs99Ptile', 'gcDurUsMedian', 'heapSizeLimit', 'heapTotal',
    'heapUsed', 'load15m', 'load1m', 'load5m', 'loopAvgTasks',
    'loopEstimatedLag', 'loopIdlePercent', 'loopsPerSecond', 'rss',
    'totalAvailableSize', 'totalHeapSizeExecutable', 'totalPhysicalSize',
    'httpClientMedian', 'httpClient99Ptile', 'httpServerMedian',
    'httpServer99Ptile', 'dnsMedian', 'dns99Ptile'
  ],
  max: [
    'gcCount', 'gcForcedCount', 'gcFullCount', 'gcMajorCount', 'systemUptime',
    'uptime', 'blockInputOpCount', 'blockOutputOpCount',
    'ctxSwitchInvoluntaryCount', 'ctxSwitchVoluntaryCount', 'ipcReceivedCount',
    'ipcSentCount', 'loopTotalCount', 'pageFaultHardCount',
    'pageFaultSoftCount', 'signalCount', 'swapCount', 'httpClientCount',
    'httpServerCount', 'dnsCount', 'httpClientAbortCount',
    'httpServerAbortCount'
  ],
  last: ['app', 'hostname', 'id', 'tags', 'title', 'user']
}

const columns = []
for (let func of Object.keys(columnAggTypes)) {
  for (let name of columnAggTypes[func]) {
    columns.push({ name, func })
  }
}

var cqs = [
  {
    name: 'downsample_to_daily',
    columns,
    fromRp: 'agentData',
    toRp: 'one_minute_daily',
    window: '1m',
    tags: ['id', 'app', 'tags']
  },
  {
    name: 'downsample_to_monthly',
    columns,
    fromRp: 'agentData',
    toRp: 'one_hour_monthly',
    window: '1h',
    tags: ['id']
  }
]

function createCQs (config, cb) {
  asynclib.eachSeries(cqs, (cq, next) => {
    createCQ(config, cq, next)
  }, cb)
}

function createCQ (config, spec, cb) {
  // first, delete the CQ if it exists
  var dropQuery = `drop continuous query ${spec.name} on nsolid`
  api.query(config.influxdb.portHTTP, 'nsolid', dropQuery, err => {
    if (err) {
      Logger.error(err, `error replacing existing continuous query ${spec.name}`)
      return cb(err)
    }

    var createQuery = `CREATE CONTINUOUS QUERY ${spec.name} ON nsolid BEGIN SELECT ` +
      spec.columns.map(column => { return `${column.func}("${column.name}") as "${column.name}"` }).join(', ') +
      ` INTO "${spec.toRp}"."metrics" ` +
      ` FROM "${spec.fromRp}"."metrics" ` +
      ` GROUP BY time(${spec.window}), ` +
      spec.tags.join(', ') +
      ' END'

    api.query(config.influxdb.portHTTP, 'nsolid', createQuery, err => {
      if (err) {
        Logger.error(err, `error creating continuous query ${spec.name}`)
        return cb(err)
      }
      cb()
    })
  })
}

function waitForInfluxDB (config, influxDBProcess, cb) {
  if (!influxDBProcess.isRunning()) {
    return cb(new Error('InfluxDB process terminated'))
  }

  if (config.influxdb.connectAttempt == null) {
    config.influxdb.connectAttempt = 0
  }

  config.influxdb.connectAttempt++
  if (config.influxdb.connectAttempt > 100) {
    return cb(new Error('unable to connect to InfluxDB'))
  }

  setTimeout(ping, 250)

  function ping () {
    Logger.debug('pinging InfluxDB to see if it\'s ready')
    api.ping(config.influxdb.portHTTP, onPing)
  }

  function onPing (err, version) {
    if (err) return waitForInfluxDB(config, influxDBProcess, cb)

    Logger.info(`InfluxDB version ${version} ready`)
    cb(null, version)
  }
}

// Launch the InfluxDB process.
function launchInfluxDB (env, logFD) {
  const platArch = `${process.platform}-${process.arch}`
  const influxPath = path.resolve(path.join(__dirname, '..', '..', 'tsdb', platArch))
  const influxExe = path.join(influxPath, 'influxd')

  env = lodash.defaults({}, env, process.env)

  const opts = {
    env: env,
    stdio: ['ignore', 'ignore', logFD]
  }

  return childProcess.spawn(influxExe, opts)
}

// This class manages the InfluxDB process.
class InfluxDBProcess {
  constructor (process, config) {
    this.process = process
    this.config = config
    this.port = config.influxdb.portHTTP
    this.url = `http://0.0.0.0:${this.port}`

    this.process.on('error', err => this._onError(err))
    this.process.on('exit', (code, signal) => this._onExit(code, signal))
  }

  isRunning () {
    return this.process != null
  }

  kill (cb) {
    if (!this.isRunning()) return

    if (cb) this.process.on('exit', cb)
    this.process.kill()
  }

  _onError (err) {
    Logger.error(err)
    this.process = null
  }

  _onExit (code, signal) {
    Logger.info(`InfluxDB exit(${code}) signal: ${signal}`)
    this.process = null
  }
}
