'use strict'

const lodash = require('lodash')

const api = require('./api')
const util = require('../util')

exports.create = create

// Buffer writes for this many millis.
const BUFFER_FOR_MS = 250
const CHUNK_SIZE = 100

const Logger = util.getLogger(__dirname, __filename)

// Create and return a new database writer.
function create (context) {
  return new Writer(context)
}

class Writer {
  constructor (context) {
    this.context = context
    this.port = this.context.config.influxdb.portHTTP
    this.timeout = null
    this.queue = []
    this.writing = false
  }

  // Write an info record to the database.
  writeInfo (agent, message) {
    // Ensure sane shape of body before modifying it
    const data = lodash.defaults({}, message.body)
    data.id = agent.id

    if (data.tags) {
      Logger.debug(`data.tags: ${JSON.stringify(data.tags)}`)
      if (lodash.isString(data.tags)) {
        data.tags = `,${data.tags},`
      } else if (lodash.isArray(data.tags)) {
        data.tags = `,${data.tags.join(',')},`
      } else {
        Logger.warn(`invalid type of tags: ${JSON.stringify(data.tags)}`)
        delete data.tags
      }
    }
    message.body = data

    // There should be a processStart field, so use that as the influx
    // time-stamp, which should be in ns instead of ms.
    if (data.processStart != null) {
      message.timeNS = `${data.processStart}000000` // big nums! don't use math!
    }

    const line = this._createLine(agent, 'info', message)
    this._write(line)
  }

  // Write an metrics record to the database.
  writeMetrics (agent, message) {
    // Ensure sane shape of body before modifying it
    const data = lodash.defaults({}, message.body)
    data.id = agent.id

    // remove nulls
    message.body = lodash(data).omitBy(lodash.isNull).value()
    const line = this._createLine(agent, 'metrics', message)
    this._write(line)
  }

  // Create a line to write to influxd
  // agent.tags should contain the influx tags, as an object map of string:string
  // message.body should contain the influx fields, as an object map of string:string
  _createLine (agent, measurement, message) {
    const tagsObject = Object.assign({}, agent.tags)
    const fieldsObject = Object.assign({}, message.body)
    this._fixupWrittenTagsAndFields(tagsObject, fieldsObject, agent, message)

    const tags = Object.keys(tagsObject).length
      ? ',' + lodash.map(tagsObject, asKeyValue).join(',')
      : ''

    const fields = lodash.map(fieldsObject, asFieldValue).join(',')

    const timestamp = message.timeNS || util.uniqDateNS()
    const influxWriteLine = `${measurement}${tags} ${fields} ${timestamp}`
    Logger.debug(`writing line to influx: ${influxWriteLine}`)
    return influxWriteLine

    function asKeyValue (val, key) {
      return `${escapeKey(key)}=${escapeKey(val)}`
    }

    function asFieldValue (val, key) {
      return `${escapeKey(key)}=${escapeFieldValue(val)}`
    }
  }

  // Sometimes the tags and fields get mixed up, fix here
  _fixupWrittenTagsAndFields (tags, fields, agent, message) {
    // Logger.info(`fixup tags and fields before: ${JSON.stringify({tags, fields}, null, 4)}`)
    agent = agent || {}
    agent.info = agent.info || {}
    message = message || {}
    message.body = message.body || {}

    // ensure all influx tags in place
    if (tags.id == null) tags.id = this._findValue('id', tags, fields, agent, message)
    if (tags.app == null) tags.app = this._findValue('app', tags, fields, agent, message)
    if (tags.hostname == null) tags.hostname = this._findValue('hostname', tags, fields, agent, message)
    if (tags.tags == null) tags.tags = this._findValue('tags', tags, fields, agent, message)

    if (Array.isArray(tags.tags)) {
      tags.tags.sort()
      tags.tags = `,${tags.tags.join(',')},`
    }

    // ensure all influx tags are not in fields
    delete fields.id
    delete fields.app
    delete fields.hostname
    delete fields.tags
    // Logger.info(`fixup tags and fields after: ${JSON.stringify({tags, fields}, null, 4)}`)
  }

  // find a property value in a few different places
  _findValue (property, tags, fields, agent, message) {
    if (tags[property] != null) return tags[property]
    if (fields[property] != null) return fields[property]
    if (agent.info[property] != null) return agent.info[property]
    if (message[property] != null) return message[property]
    if (message.body[property] != null) return message.body[property]
    return property === 'tags' ? [] : ''
  }

  // Write a line to the specified database.
  _write (line) {
    Logger.debug(`queuing line to InfluxDB: ${line}`)
    this.queue.push(line)

    if (this.queue.length >= CHUNK_SIZE) {
      this._writeQueued()
      return
    }

    if (this.timeout) return

    this.timeout = setTimeout(() => this._writeQueued(), BUFFER_FOR_MS)
    this.timeout.unref()
  }

  _writeQueued () {
    if (this.queue.length === 0) return

    // Only write one chunk at a time.  If an extended buffering period has
    // been in place, there could be many thousands of lines to process, so
    // don't stampede.
    if (this.writing) return
    this.writing = true

    Logger.debug(`writing ${this.queue.length} queued line to InfluxDB`)

    // Write the first 100 lines in the queue
    const joinedLines = this.queue.slice(0, CHUNK_SIZE).join('\n')
    this.queue = this.queue.slice(CHUNK_SIZE)

    api.write(this.port, 'nsolid', joinedLines, err => {
      this.writing = false
      if (err) {
        Logger.error(err, 'error while writing metrics to Influx, data discarded')
      }
      if (this.queue.length >= CHUNK_SIZE) {
        this._writeQueued()
      }
    })

    // Do this synchronously because data can arrive while influx is writing
    // and we want to allow a new timer to start in the meantime.
    if (this.timeout) clearTimeout(this.timeout)
    this.timeout = null
  }
}

// influx keys and tags must escape "=", " ", and ","
function escapeKey (key) {
  return key.replace(/([ =,])/g, '\\$1')
}

// format influx field values
function escapeFieldValue (value) {
  if (value == null) return '"null"'  // return null as the string 'null'
  if (lodash.isNumber(value)) return value
  if (lodash.isBoolean(value)) return `${value}`

  // if it smells like a string, ensure it's a string
  if (lodash.isString(value)) {
    value = `${value}`
  // last chance - it's an object, so JSON stringify
  } else {
    value = JSON.stringify(value)
  }

  // return quoted string, escaping quotes
  return `"${value.replace(/"/g, '\\"')}"`
}
