'use strict'

const crypto = require('crypto')
const lodash = require('lodash')

exports.initialize = initialize
exports.hash = hash
exports.getCachedPackages = getCachedPackages

// number of minutes between garbage collecting unused packages
const GC_INTERVAL_MINS = 30

// initialize
function initialize (context) {
  setInterval(() => gc(context), GC_INTERVAL_MINS * 60 * 1000).unref()
}

// return a hash of the packages data
function hash (packages) {
  packages = lodash.cloneDeep(packages)
  const hash = crypto.createHash('sha1')

  // Arrays of different orders should be considered similar enough.
  // These arrays consist of elements that are either a string or an object.
  // We can't sort on the path bc we don't want it to affect the shasum.
  // Sort on name@version or shasum.
  packages = lodash.sortBy(packages, item => `${item.name}@${item.version}`)

  packages.forEach(pkg => {
    delete pkg.path
  })

  const string = stringify(packages)
  hash.update(string)
  return hash.digest('hex')
}

// cache of packages keyed off the hash
const CachedPackages = new Map()

// given a hash and the full packages object, return a cached version of the packages
function getCachedPackages (hash, packages) {
  const cachedPackages = CachedPackages.get(hash)
  if (cachedPackages != null) return cachedPackages

  CachedPackages.set(hash, packages)
  return packages
}

// clean up unused cached packages
function gc (context) {
  const agents = context.agentManager.list()
  const seenHashes = {}

  // get all the hashes of the currently living packages
  for (let agent of agents) {
    const hash = agent.packagesHash
    seenHashes[hash] = true
  }

  // look through the existing cached packages, delete those that aren't in use
  for (let hash of CachedPackages.keys()) {
    if (seenHashes[hash] === true) continue

    CachedPackages.delete(hash)
  }
}

// stringify that sorts object keys
function stringify (obj) {
  let string = ''
  Object.keys(obj).sort().forEach(key => {
    const value = obj[key]
    if (value === null) {
      string += 'null'
    } else if (lodash.isArray(value)) {
      string += value.join(',')
    } else if (typeof value === 'object') {
      string += stringify(value)
    }
    string += value.toString()
  })
  return string
}
