'use strict'

// const request = require('client-request')
const util = require('./util')
const Logger = util.getLogger(__dirname, __filename)
const path = require('path')
const crypto = require('crypto')
const retry = require('./retry-request')
const querystring = require('querystring')
const LicenseKey = require('./license-key')
const EventEmitter = require('events')

let provider
// Number of minutes to cache license from license service
const LICENSE_CHECK_MINS = 30

module.exports.create = create

function create (context) {
  const licenseFileName = getLicenseFileName(context)
  const licenseInfo = getLicenseFileContents(licenseFileName)

  let configProvider = null

  // always set in config, value of --provider cli option, lowercased;
  // if not passed on cli, will be "default"
  let cliProvider = context.config.license.provider

  // pick provider in license.json over cli option
  if (licenseInfo && licenseInfo.provider) {
    configProvider = licenseInfo.provider
  } else {
    configProvider = cliProvider
  }

  configProvider = configProvider.toLowerCase()

  // if provider was specified in config (CLI option), but we ended up
  // using the provider in the license file, issue a warning
  if (configProvider !== cliProvider) {
    Logger.warn(`using license.json provider "${configProvider}" instead of requested provider "${cliProvider}"`)
  }

  // already have that provider, so return it
  if (provider && provider.type === configProvider) {
    return provider
  }

  // create a new provider
  if (configProvider === 'azure') {
    provider = new AzureProvider({ context: context })
  } else if (configProvider === 'aws') {
    provider = new AWSProvider({ context: context })
  } else if (configProvider === 'gcp') {
    provider = new GCPProvider({ context: context })
  } else if (configProvider === 'default') {
    provider = new DefaultProvider({ context: context })
  } else {
    throw new Error('Provider misconfigured')
  }

  return provider
}

function getLicenseFileName (context) {
  return path.join(context.config.data.dir, 'license.json')
}

function getLicenseFileContents (fileName) {
  try {
    return require(fileName)
  } catch (err) {
    return null
  }
}

// EventEmitter only used for tests.
class BaseProvider extends EventEmitter {
  constructor (params) {
    super()
    this._context = params.context
    this.canOptOut = LicenseKey.GLOBAL_OPT_OUT
    this.keyViaEmail = false
    this.filepath = getLicenseFileName(this._context)
    this.url = this._context.isStaging
      ? 'https://staging-services.nodesource.io/v2/licenses'
      : 'https://services.nodesource.io/v2/licenses'

    this.timer = null
    this.ready = null
  }

  init () {
    // attempt to load key from config file first
    let licenseKey = null
    if (this._context.config.license != null) {
      licenseKey = this._context.config.license.key
    }

    if (licenseKey != null && licenseKey !== '') {
      this.license = new LicenseKey(this.type, {key: licenseKey})
    } else {
      this.license = new LicenseKey(this.type)
    }

    const serialized = getLicenseFileContents(this.filepath)
    if (serialized == null) {
      Logger.info('No persisted license file on load.')
    } else {
      this.license = new LicenseKey(this.type, serialized)
      if (licenseKey != null && licenseKey !== '' && this.license.key !== licenseKey) {
        Logger.warn(`overriding license key ${licenseKey} from config file with saved key ${this.license.key} from ${this.filepath}`)
      }
    }

    if (this.isOptedOut()) {
      Logger.debug('License indicates optOut, so no license checking')
      return
    }

    this.startTimer()
    // This will test the validity of the license
    setImmediate(this.refreshLicense.bind(this))
  }

  isOptedOut () {
    if (LicenseKey.GLOBAL_OPT_OUT) return true
    return this.canOptOut && this.license && this.license.optOut
  }

  startTimer () {
    this.timer = setInterval(() => {
      this.refreshLicense()
    }, LICENSE_CHECK_MINS * 60 * 1000)
    this.timer.unref()
    Logger.debug(`Checking license every ${LICENSE_CHECK_MINS} minutes`)
  }

  saveKey (cb) {
    if (!this.canOptOut) {
      this.license.optOut = false
    }
    this.license.save(this.filepath, cb)
  }

  resetLicense () {
    // Used to force clear a license key
    this.license = new LicenseKey(this.type)
  }

  refreshLicense (cb) {
    cb = cb || function () {}
    Logger.debug('Beginning refresh license.')

    if (this.isOptedOut()) {
      if (!this.ready) {
        this.ready = true
        this.emit('ready')
      }
      return cb(null, this.license)
    }

    this.syncLicense((err, license) => {
      if (err) {
        Logger.warn(err, 'Could not refresh license.', license)
        return cb(null, this.license)
      }

      this.license = new LicenseKey(this.type, license)
      this.saveKey((err) => {
        if (err) {
          Logger.warn(err, 'Could not save refreshed license.')
        }
        if (!this.ready) {
          this.ready = true
          this.emit('ready')
        }
        return cb(null, this.license)
      })
    })
  }

  validateKey (key, cb) {
    Logger.debug(`validateKey ${key}`)
    if (this.isOptedOut()) {
      // It's unlikely that `this.license.optOut` is set here, but checking nonetheless
      Logger.warn('refusing to check key due to optOut')
      return setImmediate(() => {
        cb(new Error('cannot check key when optOut=true'))
      })
    }
    const options = {
      uri: `${this.url}/${key}`,
      method: 'GET',
      json: true
    }

    Logger.debug('Checking license.', options)
    retry(this._context, options, (err, res, body) => {
      if (err) return cb(err)
      if (body && body.errorMessage) {
        return cb(new Error(`Server error: ${body.errorMessage}`))
      }
      return cb(null, LicenseKey.fromBody(this.type, body))
    })
  }

  // fetches license from server and saves
  syncLicense (params, cb) {
    Logger.debug(`syncLicense(${JSON.stringify(params)})`)
    if (typeof params === 'function') {
      cb = params
      params = {}
    }

    if (params.key == null && params.email == null && this.license != null) {
      params.key = this.license.key
    }

    // if nothing to check, return current license
    if (!params.email && !params.key) {
      setImmediate(() => { cb(null, this.license) })
      return
    }

    let url = this.url
    const headers = {}

    if (params.key) {
      url = `${url}/${params.key}`
    } else {
      headers.Authorization = formatAuthHeaderEmail(params.email)

      const qparms = querystring.stringify({
        provider: this.type,
        providerData: params.providerData || ''
      })

      url = `${url}?${qparms}`
    }

    const options = {
      uri: url,
      method: 'GET',
      json: true,
      headers: headers
    }

    Logger.debug('Fetching license.', options)
    retry(this._context, options, (err, res, body) => {
      if (err) return cb(err)
      if (body && body.errorMessage) {
        return cb(new Error(`Server error: ${body.errorMessage}`))
      }
      if (res.statusCode === 404) {
        // 404 means successful lookup but no license found by this key
        this.license = new LicenseKey(this.type, {provider: this.type, key: this.license.key})
        this.saveKey(cb)
        return
      }
      if (res.statusCode !== 200) {
        return cb(new Error(`http status ${res.statusCode} fetching license`))
      }

      this.license = LicenseKey.fromBody(this.type, body)
      this.saveKey(cb)
    })
  }

  setOptOut (optOut, cb) {
    if (!this.canOptOut) {
      // not allowed to optOut
      if (optOut) {
        Logger.warn(`Ignoring optOut request for provider ${this.type}`)
        optOut = false
      }
      if (this.license.optOut) {
        Logger.warn(`Overriding optOut back to false for provider ${this.type}`)
        this.license.optOut = false
        return this.saveKey(cb)
      }
    }
    if (this.license.optOut === optOut) {
      // nothing to do...
      return setImmediate(() => {
        cb(null, this.license)
      })
    }
    this.license.optOut = optOut
    this.saveKey(cb)
  }

  createLicense (params, cb) {
    Logger.debug(`createLicense(${JSON.stringify(params)})`)

    if (!params.email) {
      setImmediate(() => {
        cb(new Error('Parameter `email` is required to create a license'))
      })
      return
    }

    if (params.providerData == null) {
      params.providerData = ''
    }

    const options = {
      uri: this.url,
      method: 'POST',
      json: true,
      body: params,
      headers: {
        Authorization: formatAuthHeaderCreator(this._context.version)
      }
    }

    retry(this._context, options, (err, res, body) => {
      if (err) return cb(err)
      if (res.statusCode < 200 || res.statusCode > 202) return cb(new Error(res.statusCode))
      if (body == null) return cb(new Error('no body returned from createLicense'))
      if (body.errorMessage) {
        return cb(new Error(`Server error: ${body.errorMessage}`))
      }

      if (this.keyViaEmail) {
        this.resetLicense()
      } else {
        this.license = LicenseKey.fromBody(this.type, body)
      }
      this.saveKey(cb)
    })
  }
}

class DefaultProvider extends BaseProvider {
  constructor (params) {
    super(params)
    this.type = 'default'
    this.keyViaEmail = true
    this.init()
  }
}

class AWSProvider extends BaseProvider {
  constructor (params) {
    super(params)
    this.type = 'aws'
    this.canOptOut = true
    this.init()
  }
}

class AzureProvider extends BaseProvider {
  constructor (params) {
    super(params)
    this.type = 'azure'
    this.init()
  }
}

class GCPProvider extends BaseProvider {
  constructor (params) {
    super(params)
    this.type = 'gcp'
    this.canOptOut = true
    this.init()
  }
}

function formatAuthHeaderEmail (email) {
  const hashedEmail = crypto.createHash('sha256')
    .update(email)
    .digest('hex')

  const authString = Buffer.from(`${hashedEmail}:NodeSource`).toString('base64')
  return `Basic ${authString}`
}

function formatAuthHeaderCreator (version) {
  const authString = Buffer.from(`nsolid-storage ${version}:NodeSource`).toString('base64')
  return `Basic ${authString}`
}
