'use strict'

module.exports = checkWithRetries

const request = require('client-request')

const util = require('./util')
const Logger = util.getLogger(__dirname, __filename)
const lodash = require('lodash')

const RETRY_ATTEMPTS = 5         // # of retry attempts
const REQUEST_TIMEOUT_MS = 15000  // # of ms to wait before timing out
const RETRY_MIN_MS = 200         // min amount of time to wait before retry
const RETRY_RANGE_MS = 500       // max amount of random time to add to retry

// run the request, perhaps multiple times, w/backoff
// Options:
//   uri      string
//   method   GET, POST, PUT, et. al.
//   json     boolean, whether response body will be JSON
//   body     string or object
//   timeout  number
//   headers  object
function checkWithRetries (context, options, cb) {
  let attempt = 1
  options = lodash.defaults({
    retryAttempts: RETRY_ATTEMPTS,
    timeout: REQUEST_TIMEOUT_MS,
    retryMin: RETRY_MIN_MS,
    retryRange: RETRY_RANGE_MS
  }, options)
  const retryAttempts = options.retryAttempts
  delete options.retryAttempts

  run(context, options, getHandler(retry))

  function retry (err, res, body) {
    if (!err) return cb(null, res, body)

    attempt++
    if (attempt > retryAttempts) {
      Logger.error(err, 'request retries exceeded')
      return cb(err)
    }

    const backoff = options.retryMin + Math.floor(Math.random() * options.retryRange)
    delete options.retryMin
    delete options.retryRange

    setTimeout(function () {
      run(context, options, getHandler(retry))
    }, backoff).unref()
    Logger.debug(err, 'retry in %s ms', backoff)
  }
}

function getHandler (cb) {
  return function (err, res, body) {
    // send back err if that's what we got
    if (err) return cb(err)

    // no result is an error condition
    if (res == null) {
      return cb(new Error('unexpected null result'))
    }

    // no HTTP status is an error condition
    if (res.statusCode == null) {
      return cb(new Error('unexpected missing http status'))
    }

    cb(null, res, body)
  }
}

// Run an HTTP request and return the response.
// cb called with (err, result)
function run (context, options, cb) {
  cb = util.onlyCallOnce(cb)
  const fetchStart = Date.now()

  options.agent = context.httpsAgent

  const req = request(options, handleReply)
  Logger.debug(`request: ${JSON.stringify(options)}`)
  req.on('error', cb)

  function handleReply (err, res, body) {
    const secs = (Date.now() - fetchStart) / 1000
    const method = options.method
    const status = res.statusCode
    Logger.debug(`response: ${status} ${method} ${options.uri} ${secs}s ${JSON.stringify(body)}`)

    return cb(err, res, body)
  }
}
