'use strict'

const os = require('os')
const v8 = require('v8')

const lodash = require('lodash')

const util = require('./util')

exports.getInfo = getInfo
exports.getMetrics = getMetrics
exports.getPackages = getPackages

const Logger = util.getLogger(__dirname, __filename)

Logger.debug('startup')

const Metrics = new Map()

function getPackages (id) {
  return lodash.clone(examplePackages)
}

// Return a simulated metrics object.
function getMetrics (id) {
  let metrics = Metrics.get(id)
  if (metrics == null) {
    metrics = initMetrics()
    Metrics.set(id, metrics)
  }

  return genMetrics(id, metrics)
}

function genMetrics (id, metrics) {
  metrics = lodash.defaults({}, metrics)

  for (let key in metrics) {
    if (typeof metrics[key] === 'string') continue
    metrics[key] = metrics[key]()
  }

  metrics.time = Date.now()

  Logger.debug(`getting metrics for ${id}: ${JSON.stringify(metrics)}`)

  return metrics
}

function initMetrics () {
  const memoryUsage = process.memoryUsage()
  const heapStats = v8.getHeapStatistics()
  const loadAvg = os.loadavg()

  return {
    uptime: getRandomWalker(process.uptime()),
    rss: getRandomWalker(memoryUsage.rss),
    heapTotal: getRandomWalker(memoryUsage.heapTotal),
    heapUsed: getRandomWalker(memoryUsage.heapUsed),
    activeRequests: getRandomWalker(process._getActiveRequests().length),
    activeHandles: getRandomWalker(process._getActiveHandles().length),
    title: 'sim-title',
    user: 'sim-user',
    cpu: getRandomWalker(50),
    totalHeapSizeExecutable: getRandomWalker(heapStats.total_heap_size_executable),
    totalPhysicalSize: getRandomWalker(heapStats.total_physical_size),
    totalAvailableSize: getRandomWalker(heapStats.total_available_size),
    heapSizeLimit: getRandomWalker(heapStats.heap_size_limit),

    freeMem: getRandomWalker(os.freemem()),
    systemUptime: getRandomWalker(os.uptime()),
    load1m: getRandomWalker(loadAvg[0], true),
    load5m: getRandomWalker(loadAvg[1], true),
    load15m: getRandomWalker(loadAvg[2], true),
    cpuSpeed: getRandomWalker(os.cpus()[0].speed)
  }
}

// Return a simulated info object.
function getInfo (id) {
  return {
    id: id,
    app: appCycler(),
    pid: (Math.random() * 65536) | 0,
    hostname: hostCycler(),
    tags: [tag1Cycler(), tag2Cycler()],
    versions: {
      http_parser: '2.7.0',
      node: '6.2.1',
      nsolid: '2.0.0'
    },
    execPath: '/usr/local/bin/nsolid',
    main: '/Users/sim/app',
    arch: 'sim-arch',
    platform: 'sim-plat',
    totalMem: 16 * 1024 * 1024 * 1024,
    cpuCores: 8,
    cpuModel: 'sim-cpu',
    processStart: Date.now()
  }
}

// Build a bunch of cycled value getters.
const appCycler = getCycler(['front-end', 'middle-out', 'back-end'])
const hostCycler = getCycler(['host-a.example.com', 'host-b.example.com'])
const tag1Cycler = getCycler(['region:west', 'region:east', 'region:ap'])
const tag2Cycler = getCycler(['env:staging', 'env:production'])

// Build a cycled value getter.
function getCycler (values) {
  let current = -1

  return function getNext () {
    current++
    if (current >= values.length) current = 0

    return values[current]
  }
}

// Build a random walker between 0.8 and 1.2 of init value
function getRandomWalker (value, isFloat) {
  const min = 0.25 * value
  const max = 4.0 * value
  const step = (max - min) / 100
  let current = value

  return getNext

  function getNext () {
    current += Math.random() < 0.5 ? -step : step
    if (current < min) current = min
    if (current > max) current = max

    if (isFloat) return current

    return Math.round(current)
  }
}

const examplePackages = {
  'packages': [{
    'dependencies': [
      'node_modules/async',
      'node_modules/express'
    ],
    'main': 'use-resources.js',
    'path': '/home/erin/use-resources',
    'version': '1.0.0',
    'name': 'use-resources',
    'modules': []
  }, {
    'path': '/home/erin/use-resources/node_modules/async',
    'version': '2.0.1',
    'dependencies': [],
    'main': 'dist/async.js',
    'name': 'async',
    'modules': [{
      'path': 'dist/async.js',
      'shasum': 'b9da1e052d92204a0f54b7d37e7a864b8c4676cd'
    }]
  }, {
    'name': 'call',
    'modules': [{
      'shasum': '41d5d3a6247d29c2e9723442f87bdfad60d9abbf',
      'path': 'index.js'
    }],
    'version': '3.0.1',
    'path': '/home/erin/use-resources/node_modules/call',
    'main': 'index.js',
    'dependencies': []
  }, {
    'modules': [{
      'path': 'index.js',
      'shasum': 'e10c187c36324757d21467047af86fce9b06ad86'
    }],
    'name': 'express',
    'version': '3.1.2',
    'path': '/home/erin/use-resources/node_modules/express',
    'main': 'index.js',
    'dependencies': []
  }]
}
