'use strict'

const lodash = require('lodash')

const Cache = require('./thresholds/cache')
const util = require('./util')

const Logger = util.getLogger(__dirname, __filename)

const THRESHOLD_CONFIG_SETTINGS_KEY = 'thresholds'
const CONFIGURED_HEAP_LIMIT = 1.5
const THRESHOLDS_PROFILE_DURATION = 5 // in seconds

var thresholdConfig = {}

function setThresholdConfig (context, config) {
  const settingsManager = context.settingsManager

  const settings = settingsManager.getSettings()
  settings[THRESHOLD_CONFIG_SETTINGS_KEY] = lodash.cloneDeep(config)

  settingsManager.setSettings(settings, (err) => {
    if (err) {
      Logger.error(err, 'error setting thresholds config')
    }
  })

  Object.keys(config).forEach(key => {
    thresholdConfig[key] = config[key]
  })
}

function getThresholdConfig (context) {
  const settings = getSettings(context)
  const config = settings[THRESHOLD_CONFIG_SETTINGS_KEY] || {}
  return config
}

const defaultConfig = {
  license: {
    key: null
  },
  thresholdActionConfig: {
    profileDuration: THRESHOLDS_PROFILE_DURATION
  },
  thresholds: {
    app: thresholdConfig
  }
}

exports.cache = new Cache(defaultConfig)

function initThresholdActions (context) {
  // set up config based on context, initializing populating from saved settings
  setThresholdConfig(context, getThresholdConfig(context))
  const config = lodash.defaultsDeep({
    httpsAgent: context.httpsAgent,
    license: context.provider.license,
    context: context,
    notificationEndpoint: {
      url: context.isStaging
        ? 'https://staging-services.nodesource.io/notification'
        : 'https://services.nodesource.io/notification'
    },
    httpBaseUrl: util.getHttpBaseURL(context)
  }, defaultConfig)

  exports.cache = new Cache(config, exports.cache, context)

  setInterval(function cycle () {
    config.httpsAgent = context.httpsAgent
    config.thresholds.app = thresholdConfig
    config.license = context.provider.license

    exports.cache = new Cache(config, exports.cache, context)
  }, 1000).unref()

  return exports.cache
}

const GB = 1024 * 1024 * 1024

function addProcess (metrics, processInfo) {
  const proc = {
    app: processInfo.app,
    id: processInfo.id,
    cpu: metrics.cpu,
    memory: {
      used: metrics.heapUsed / (CONFIGURED_HEAP_LIMIT * GB)
    },
    tags: processInfo.tags,
    pid: processInfo.pid,
    host: processInfo.hostname,
    hostname: processInfo.hostname,
    title: processInfo.execPath + ' ' + processInfo.main
  }
  exports.cache.addProcess(proc)
}

exports.initThresholdActions = initThresholdActions
exports.setThresholdConfig = setThresholdConfig
exports.getThresholdConfig = getThresholdConfig
exports.addProcess = addProcess

let CachedSettings = null

function getSettings (context) {
  // if there's already a copy of the settings, return it
  if (CachedSettings != null) return CachedSettings

  // if there's not a copy of the settings, this is the first time through:
  // 1: get the settings
  // 2: set up event handler watching for changed settings

  CachedSettings = context.settingsManager.getSettings()

  context.settingsManager.on('settings-changed', settingsChanged)

  return CachedSettings

  function settingsChanged () {
    CachedSettings = context.settingsManager.getSettings()
  }
}
