'use strict'

const async = require('async')
const url = require('url')
const path = require('path')
const doSnapshot = require('./actions/snapshot')
const doProfile = require('./actions/profile')
const doNotify = require('./actions/notify')

const util = require('../util')
const Logger = util.getLogger(__dirname, __filename)

class ThresholdAction {
  constructor (metric, value, cache, state, thresholdConfig, context, serverUrl) {
    var self = this
    this.metric = metric
    this.state = state
    this.cache = cache
    this.thresholdConfig = thresholdConfig
    this.context = context || cache.context
    this.serverUrl = serverUrl
    this.results = []

    // Capture this early in case the agent exits while we're processing
    const agent = this.context.agentManager ? this.context.agentManager.byId(state.id) : {}

    const consoleURL = this.context.config.notification.consoleURL

    var order = ['snapshot', 'profile', 'notify', 'webhook']
    var i = 0
    this.orderRev = new Map(order.map(name => { return [name, i++] }))
    this.actions = {
      snapshot: function performSnapshot (context, appConfig, actionConfig, proc, done) {
        Logger.debug('snapshot on process %s (%s)', proc.id, proc.app)
        appConfig.snapshotEndpoint = { url: serverUrl }
        doSnapshot(appConfig, proc.id, (err, result) => {
          if (err) {
            return done(err)
          }

          const heapUrl = url.resolve(consoleURL, '/' + path.join(
            'app',
            result.app,
            'heap',
            result.id,
            result.asset + ''
          ))

          done(null, {
            url: heapUrl,
            result
          })
        })
      },
      profile: function performProfile (context, appConfig, actionConfig, proc, done) {
        const duration = actionConfig.profileDuration
        Logger.debug('profile on process %s (%s) for %s', proc.id, proc.app, duration)
        appConfig.profileEndpoint = { url: serverUrl }
        doProfile(appConfig, proc.id, duration, (err, result) => {
          if (err) {
            return done(err)
          }

          const profileUrl = url.resolve(consoleURL, '/' + path.join(
            'app',
            result.app,
            'profile',
            result.id,
            result.asset + '',
            'flamegraph'
          ))

          done(null, {
            url: profileUrl,
            result
          })
        })
      },
      notify: function performNotify (context, appConfig, actionConfig, proc, done) {
        if (context.provider.isOptedOut()) {
          Logger.debug('Skipping notify due to opt-out.')
          return done()
        }

        const payload = {
          type: 'threshold',
          method: 'email',
          data: {
            metric: metric,
            value: value,
            threshold: thresholdConfig.threshold,
            recipients: actionConfig.recipients,
            delay: thresholdConfig.delay,
            delayUnits: thresholdConfig.delayUnits,
            process: proc,
            artifacts: self.results
          }
        }

        Logger.debug('notify on process %s (%s)', proc.id, proc.app)
        appConfig.notificationEndpoint = {
          url: context.isStaging
            ? 'https://staging-services.nodesource.io/notification'
            : 'https://services.nodesource.io/notification'
        }

        appConfig.license = {
          key: cache.config.license.key
        }

        doNotify(appConfig, payload, done)
      },
      webhook: function performWebhook (context, appConfig, actionConfig, proc, done) {
        const threshold = {}
        threshold[metric] = {
          value: value,
          enabled: thresholdConfig.enabled,
          threshold: thresholdConfig.threshold,
          delay: thresholdConfig.delay,
          delayUnits: thresholdConfig.delayUnits,
          snooze: thresholdConfig.snooze,
          snoozeUnits: thresholdConfig.snoozeUnits,
          cpu: proc.cpu,
          memory: proc.memory
        }

        var assets = {}
        for (let res of self.results) {
          if (res.url) {
            assets[res.type] = res.url
          }
        }
        const params = {
          metric,
          agent,
          threshold,
          assets
        }

        context.webhooksManager.sendThresholdExceeded(params, done)
      }
    }
  }

  getActions () {
    if (!this.thresholdConfig) {
      return false
    }

    const actions = this.thresholdConfig.actions
    if (!actions) {
      return false
    }

    return actions
  }

  execute (config, next) {
    var self = this
    const configuredActions = this.getActions()
    const actions = this.actions
    const metric = this.metric
    const cache = this.cache
    const state = this.state
    const context = this.context

    const processid = state.id

    if (!configuredActions) {
      return finish()
    }

    const actionNames = Object.keys(configuredActions).sort((key1, key2) => {
      // use -1 for unknown actions, so they are run first
      return (this.orderRev.get(key1) || -1) - (this.orderRev.get(key2) || -1)
    })

    async.eachSeries(actionNames, processAction, finish)

    function processAction (actionName, done) {
      const action = actions[actionName]
      const actionConfig = configuredActions[actionName]

      if (!validateActionConfig(action, actionConfig)) {
        return done()
      }

      Logger.info(`performing ${actionName} for app ${state.app} ${state.id}`)
      action(context, config, actionConfig, state, (err, data) => {
        data = data || {}
        if (err) {
          Logger.error(err, `error performing action ${actionName}: ${err.message} ${err.code}`)
          // Don't propagate error to async because we don't want to interfere
          // with other actions
          self.results.push({
            error: {
              code: err.code,
              message: err.message
            }
          })
          return done()
        }

        // actions that produce an artifact url will be passed to notifications and webhooks
        self.results.push({
          type: actionName,
          result: data.result,
          url: data.url
        })

        done()
      })
    }

    function finish (err) {
      const proc = cache.getProcess(processid)

      // if the process crashed, avoid exploding
      if (!proc) {
        return next()
      }

      if (!proc.thresholds) {
        proc.thresholds = {}
      }
      proc.thresholds[metric] = { snoozeStart: Date.now() }
      next(err, self.results)
    }
  }
}

module.exports = ThresholdAction

// Return true if any part of the config for this action exists and is valid
// and enabled, else, return false.
function validateActionConfig (action, actionConfig) {
  if (Array.isArray(actionConfig)) {
    if (!action || !actionConfig) {
      return false
    }
    // If every entry is not disabled, return true
    return !actionConfig.every(entry => !entry.enabled)
  } else {
    if (!action || !actionConfig || !actionConfig.enabled) {
      return false
    }
    return true
  }
}
