'use strict'

const request = require('client-request')
const util = require('./util')
const Logger = util.getLogger(__dirname, __filename)
const path = require('path')
const fs = require('fs')
const hasher = require('./package-hasher')
const semver = require('semver')
const asynclib = require('async')
const sendEmail = require('./thresholds/actions/notify')
const lodash = require('lodash')
const cache = require('./threshold-actions').cache

const snykReportURL = 'https://snyk.io/partners/api/v2/vulndb/npm.json?token=nodesource-bf18e772-0f23-4154-8905-5606cc2757e1'
const emailCache = {}

exports.create = create
exports.snykReportURL = snykReportURL

function create (context) {
  return new VulnManager(context)
}

class VulnManager {
  constructor (context) {
    this._context = context
    this.filepath = path.join(context.config.data.dir, 'snyk-report.json')
    this.vulnsCache = new Map()
    this.list = require('./snyk-snapshot.json').npm
    this._isNotifying = {}

    // read the cached file, if available
    this._readVulnDB()

    setImmediate(() => {
      this._runSnykCheck()
    })

    // Default is 30 minutes; if set to 0, reset to 30 min instead
    let refreshMins = context.config.vulnerabilities.refreshMinutes
    if (refreshMins === 0) refreshMins = 30
    Logger.debug(`Checking vulnerabilities list every ${refreshMins} minutes`)

    setInterval(() => {
      this._runSnykCheck()
    }, refreshMins * 60 * 1000).unref()

    const cacheCleanMinutes = 30
    setInterval(() => {
      this._cleanEmailCache()
    }, cacheCleanMinutes * 60 * 1000).unref()
  }

  list () {
    return this.vulnsCache
  }

  _runSnykCheck () {
    if (!this._checkVulnList) {
      // No _checkVulnList happens when the instance has been garbage
      // collected but the timer is still firing. Only pertinent to
      // tests and shutdown.
      return
    }

    Logger.debug('Refreshing vulnerability list.')
    this._checkVulnList(isCacheHit => {
      if (isCacheHit) {
        return
      }
      this._downloadVulnList(err => {
        if (!err) return

        Logger.warn(err, 'Did not download new vulnerability list.')
      })
    })
  }

  isVulnPackageMatch (vuln, pkg) {
    if (!vuln || !pkg || !vuln.packageName || !pkg.name || !vuln.semver) {
      return false
    }

    if (vuln.packageName !== pkg.name ||
      !semver.satisfies(pkg.version, vuln.semver.vulnerable)) {
      return false
    }

    // see if the vuln has been patched
    if (vuln.id && pkg.flagFiles) {
      // Snyk vuln id's use colon, but files have hyphens, so xlate -> hyphen.
      // Can't translate the other way, since hyphen is valid in a package name.
      // Example flag file names:
      //   .snyk-npm-marked-20140131-1.flag   => id: npm:marked:20140131-1
      //   .snyk-npm:npm:marked:20150520.flag => id: npm:marked:20150520
      const vulnId = vuln.id.replace(/:/g, '-')
      for (let flagFile of pkg.flagFiles) {
        // flag file for this vuln, so return false
        if (flagFile.endsWith(`${vulnId}.flag`)) {
          return false
        }
      }
    }

    // satisfies vulnerable check, no flag files => it's vulnerable
    return true
  }

  _notifySoon (vuln) {
    if (vuln == null || vuln.length === 0) return

    // This happens when the agents are running when storage service starts
    if (this._isNotifying[vuln.id]) {
      return
    }

    const vulnerabilities = []
    vuln.forEach((report) => {
      var vulnerable = lodash.get(report, 'semver.vulnerable', '*')
      var unaffected = lodash.get(report, 'semver.unaffected', 'N/A')
      // NB: Most of these properties are ignored right now
      //     but this is what was sent in 1.4.
      vulnerabilities.push({
        id: report.id,
        package: report.packageName,
        title: report.title,
        published: report.publicationTime,
        credit: report.credit,
        ids: report.identifiers,
        vulnerable: vulnerable,
        unaffected: unaffected,
        description: report.description
      })
    })

    if (!vulnerabilities.length) {
      return
    }

    Logger.debug('about to send a report with this listing:')
    vulnerabilities.forEach((report) => {
      Logger.debug(`${report.package}@${report.vulnerable} - ${report.title}`)
    })

    const agents = this._context.agentManager.byPackages(vuln.hash)

    if (!agents.length) {
      // This happens when we get the list of packages from the agent
      // before we get the agent's info message. Wait until the next
      // iteration of vuln checking to send emails. We'll have the
      // agent info by then and be able to save the fact that we sent emails.
      // I doubt this will happen since we moved notifySoon to run on the
      // next tick of the event loop.
      return
    }

    const tasks = this._doWebhooks(agents, vuln, vulnerabilities)
    const fn = this._doEmail(agents, vuln, vulnerabilities)
    if (fn) tasks.push(fn)
    if (!tasks.length) return

    this._isNotifying[vuln.id] = true
    asynclib.parallel(tasks, err => {
      if (err) Logger.error('Error sending emails or webhooks', err)
      delete this._isNotifying[vuln.id]
    })
  }

  // Return an array of async.parallel tasks
  _doWebhooks (agents, vuln, vulnerabilities) {
    const settings = this._context.settingsManager.getSettings()
    const self = this
    let webhook = lodash.get(settings, 'security.actions.webhook')
    if (!webhook || (!webhook.enabled && !webhook.length)) {
      return []
    }

    // Has this webhook already seen every vulnerability
    const everyVulnWebhookIsOld = vulnerabilities.every(_vuln => {
      // Has this webhook already been notified?
      const everyWebhookIsOld = agents.every(agent => {
        return emailCache[agent.id] && emailCache[agent.id][_vuln.id] &&
          emailCache[agent.id][_vuln.id][webhook.url] === true
      })
      return everyWebhookIsOld
    })

    if (everyVulnWebhookIsOld) return []

    return vulnerabilities.map(_vuln => {
      return acb => {
        asynclib.eachLimit(agents, 3,
          (agent, parCb) => {
            emailCache[agent.id] = emailCache[agent.id] || {}
            emailCache[agent.id][_vuln.id] = emailCache[agent.id][_vuln.id] || {}
            emailCache[agent.id][_vuln.id][webhook.url] = true

            self._context.webhooksManager.sendVulnerabilityDetected({
              agent: agent,
              vulnerability: _vuln
            }, parCb)
          },
          acb
        )
      }
    })
  }

  // Return an async.parallel task or undefined
  _doEmail (agents, vuln, vulnerabilities) {
    const settings = this._context.settingsManager.getSettings()
    let recipients = lodash.get(settings, 'security.actions.notify.recipients')
    if (!Array.isArray(recipients)) {
      recipients = []
    }
    const enabled = lodash.get(settings, 'security.actions.notify.enabled')
    const isEmail = lodash.get(settings, 'security.actions.notify.method') === 'email'
    if (!enabled || !isEmail || recipients.length === 0) {
      // Nothing to do with this report.
      return
    }

    var appConfig = {
      notificationEndpoint: {
        url: this._context.isStaging
          ? 'https://staging-services.nodesource.io/notification'
          : 'https://services.nodesource.io/notification'
      },
      license: {
        key: this._context.provider.license.key
      },
      httpsAgent: this._context.httpsAgent
    }

    const payload = {
      type: 'security',
      method: 'email',
      data: {
        recipients: recipients,
        vulnerabilities: vulnerabilities
      }
    }

    // Only send notifications to recipients who haven't yet been
    // notified. This can be important if an email address was added
    // but no new vulns have been added.
    payload.data.recipients = payload.data.recipients.filter(email => {
      // Check to see if there are new vulns for this email.
      const everyVulnEmailIsOld = payload.data.vulnerabilities.every(_vuln => {
        // Check to see if there are new processes for this vuln.
        const everyEmailIsOld = agents.every(agent => {
          return emailCache[agent.id] && emailCache[agent.id][_vuln.id] &&
            emailCache[agent.id][_vuln.id][email] === true
        })
        return everyEmailIsOld
      })

      return !everyVulnEmailIsOld
    })

    // All vulns are read, skip notifying.
    if (!payload.data.recipients.length) {
      return
    }

    return acb => {
      sendEmail(appConfig, payload, (err) => {
        if (err) {
          Logger.error(err, 'error sending security email notification')
          return acb(err)
        }
        Logger.info('sent security notification emails', payload.data.recipients)

        payload.data.recipients.forEach(email => {
          payload.data.vulnerabilities.forEach(_vuln => {
            agents.forEach(agent => {
              emailCache[agent.id] = emailCache[agent.id] || {}
              emailCache[agent.id][_vuln.id] = emailCache[agent.id][_vuln.id] || {}
              emailCache[agent.id][_vuln.id][email] = true
            })
          })
        })
        acb()
      })
    }
  }

  // packages are ingested from agents with this method
  addPackages (packages) {
    packages = trimPackages(packages)
    fixFlagFileNames(packages)

    const hash = hasher.hash(packages)
    Logger.debug('got packages', hash, packages)

    if (!this.vulnsCache.has(hash)) {
      const vuln = this._generateVulnList(packages)
      vuln.hash = hash
      this.vulnsCache.set(hash, vuln)
      Logger.debug('got vuln report', hash, vuln)

      // Do this on the next tick, since hash gets assigned to agent
      // synchronously after this function call
      setImmediate(() => this._notifySoon(vuln))
    }

    return hash
  }

  rebuildVulnsCache (cb) {
    const self = this
    const list = this._context.agentManager.list()
    asynclib.reduce(list, new Map(), asynclib.ensureAsync(iteratee), callback)

    function iteratee (cache, agent, cb) {
      if (!agent.packagesHash || !agent.packages || !agent.packages.length) {
        return cb(null, cache)
      }

      if (!cache.has(agent.packagesHash)) {
        const vuln = self._generateVulnList(agent.packages)
        vuln.hash = agent.packagesHash
        cache.set(agent.packagesHash, vuln)
        self._notifySoon(vuln)
      }

      if (cb) cb(null, cache)
    }

    function callback (err, cache) {
      if (err) {
        cb(err)
        Logger.error('Could not rebuild vulns cache.', err)
        return
      }

      if (!cache.size) {
        if (cb) return cb()
        return
      }

      self.vulnsCache = cache
      if (cb) cb()
    }
  }

  /*
   * Callback boolean argument describes whether
   * the cached vulnerability list is identical.
   */
  _checkVulnList (cbIdentical) {
    if (this._context.provider.isOptedOut()) {
      Logger.debug('_checkVulnList() - early exit because optOut, re-read vuln file')
      return cbIdentical(false)
    }

    // Always download if we're just starting up, because we don't have a
    // cached value for the etag yet.
    if (!this.list) {
      setImmediate(() => {
        cbIdentical(false)
      })
      return
    }

    request({
      uri: exports.snykReportURL,
      method: 'HEAD'
    }, (err, response) => {
      if (err) {
        Logger.error('Error requesting vuln list headers', err)
        return cbIdentical(false)
      }

      const etag = response.headers['etag']
      if (etag && etag === this.etag) {
        cbIdentical(true)
      } else {
        cbIdentical(false)
      }
    })
  }

  // Read the cached vuln db, writing out current version if the file doesn't
  // exist.  Updates `this.list` upon successful re-read, otherwise does not
  // update it.
  _readVulnDB () {
    Logger.debug('_readVulnDB() - re-reading vuln db')
    let content

    try {
      content = fs.readFileSync(this.filepath, 'utf8')
    } catch (err) {
      if (err.code === 'ENOENT') {
        this._writeVulnDB()
      } else {
        Logger.warn(err, `error reading ${this.filepath}`)
      }
      return
    }

    try {
      this.list = JSON.parse(content).npm || {}
    } catch (err) {
      Logger.warn(err, `error parsing JSON content in ${this.filepath}`)
    }
  }

  _writeVulnDB () {
    let content = JSON.stringify({ npm: this.list }, null, 2)
    try {
      fs.writeFileSync(this.filepath, content, 'utf8')
    } catch (err) {
      Logger.warn(err, `error writing ${this.filepath}`)
    }
  }

  _downloadVulnList (cb) {
    cb = util.onlyCallOnce(cb)

    // if opted out, try re-reading the file
    if (this._context.provider.isOptedOut()) {
      Logger.debug('_downloadVulnList() - re-reading vuln list because optOut')
      this._readVulnDB()
      setImmediate(() => {
        this.rebuildVulnsCache()
        cb()
      })
      return
    }

    // if not opted out, download the file
    let fileContents = ''
    let etag

    const snykReportURL = exports.snykReportURL
    const httpAgent = util.getHttpAgent(this._context, snykReportURL)

    request({
      uri: exports.snykReportURL,
      agent: httpAgent,
      method: 'GET'
    }, (err, response, body) => {
      if (err) {
        Logger.error('Error requesting vuln list headers', err)
        return cb(err)
      }

      if (response.statusCode !== 200) {
        cb(new Error(`HTTP error: ${response.statusMessage}`))
        return
      }

      try {
        fileContents = JSON.parse(body)
      } catch (err) {
        cb(err)
        return
      }

      this.normalizeVulnDB(fileContents)

      let list = fileContents
      fileContents = JSON.stringify(fileContents, null, 4)
      fs.writeFile(this.filepath, fileContents, { flag: 'w' }, err => {
        if (err) {
          return cb(err)
        }

        this.list = list.npm || {}
        this.etag = etag
        setImmediate(() => {
          this.rebuildVulnsCache()
          cb()
        })
      })
    })
  }

  // convert vuln db v2 to compatible format
  normalizeVulnDB (body) {
    for (let packageName in body.npm) {
      const packageVulns = body.npm[packageName]
      for (let packageVuln of packageVulns) {
        const sver = packageVuln.semver
        if (sver == null) continue

        // Snyk changed the semver property arrays from strings to arrays in
        // the change from /api/v1 to /api/v2.  We're sticking with the string
        // version as our canonical format.
        // Snyk says the vulnerable and unaffected arrays for npm
        // will only ever have one element, but just in case they don't,
        // let's join 'em with ||, which is what I think this would logically
        // equates to.
        if (sver.vulnerable) sver.vulnerable = sver.vulnerable.join(' || ')
        if (sver.unaffected) sver.unaffected = sver.unaffected.join(' || ')
      }
    }
  }

  /*
   * This generates a per-agent vulnerability list for caching.
   * A different format is used for returning from the API
   */
  _generateVulnList (packages, vulnList) {
    if (!packages) {
      return []
    }

    vulnList = vulnList || this.list

    const filteredSet = new Set()

    return packages.reduce((filteredList, cur) => {
      if (!vulnList[cur.name]) {
        return filteredList
      }

      // add package if vulnerable
      vulnList[cur.name].forEach(vuln => {
        if (this.isVulnPackageMatch(vuln, cur)) {
          if (!filteredSet.has(vuln.id)) {
            filteredList.push(vuln)
            filteredSet.add(vuln.id)
          }
        }
      })

      return filteredList
    }, [])
  }

  /*
   * This generates a full list of all known vulnerabilities across the
   * system and the apps and processes affected.
   *
   * GET /v1/vulnerabilities
   */
  getFullVulnReport () {
    const formattedVulns = new util.SerializableMap()
    const agents = this._context.agentManager.list()

    // Report
    this.vulnsCache.forEach((report, hash) => {
      agents.forEach(agent => {
        if (agent.packagesHash !== hash) {
          return
        }

        const packages = agent.packages
        report.forEach(vuln => {
          // Apps
          packages.forEach(pkg => {
            // Determine if the package matches the vulnerability
            if (!this.isVulnPackageMatch(vuln, pkg)) {
              return
            }

            // Add the vulnerability to the response.
            let formattedVuln = getFormattedVuln(formattedVulns, vuln, pkg)
            if (!formattedVuln) return

            // Add the app to the vulnerability
            let app = getVulnerabilityApp(formattedVuln, agent)
            if (!app) return

            // Add the process to the app
            let process = getAppProcess(app, agent)
            if (!process) return

            // These will always be unique unless npm is really screwed up,
            // so we don't have to use a map here.
            process.loaded.push({
              version: pkg.version,
              path: pkg.path
            })
          })
        })
      })
    })

    return formattedVulns
  }

  /*
   * This generates a full list of all known vulnerabilities for a
   * single app and processes affected.
   *
   * GET /v1/application/:id
   */
  getVulnsAndProcesses (app) {
    const result = {
      processes: [],
      vulnerabilities: new util.SerializableMap()
    }

    // Process all the agents.
    this._context.agentManager.list().forEach((agent) => {
      if (agent.info == null) return
      if (agent.metrics == null) return
      if (agent.info.app !== app) return

      const vulns = this._context.vulnManager.getVulnsByAgent(agent)
      vulns.forEach(vuln => {
        result.vulnerabilities.set(`${vuln.packageName}:${vuln.title}`, {
          id: vuln.id,
          package: vuln.packageName,
          patches: vuln.patches,
          title: vuln.title
        })
      })

      var cachedProcess = cache.getProcess(agent.info.id)
      const proc = {
        id: agent.info.id,
        app: agent.info.app,
        hostname: agent.info.hostname,
        tags: agent.info.tags,
        cpu: agent.metrics.cpu,
        heapUsed: agent.metrics.heapUsed,
        title: agent.metrics.title,
        uptime: agent.metrics.uptime,
        pid: agent.info.pid,
        processStart: agent.info.processStart,
        // null indicates that we don't have a packages list yet
        packageCount: agent.packages ? agent.packages.length : null,
        vulnerabilityCount: vulns.length,
        thresholds: cachedProcess && cachedProcess.thresholds ? cachedProcess.thresholds : {}
      }

      result.processes.push(proc)
    })

    return result
  }

  getVulnsByAgent (agent) {
    const report = this.vulnsCache.get(agent.packagesHash)
    if (!report) {
      return []
    }

    return report
  }

  _cleanEmailCache () {
    const agents = this._context.agentManager.list()

    Object.keys(emailCache).forEach(agentId => {
      const agent = agents.find(agent => agent.id === agentId)
      if (agentId && !agent) {
        delete emailCache[agentId]
      }
    })
  }
}

function getFormattedVuln (formattedVulns, vuln, pkg) {
  if (!formattedVulns || !vuln || !vuln.id || !pkg) return

  let formattedVuln = formattedVulns.get(vuln.id)
  if (!formattedVuln) {
    formattedVuln = {
      package: vuln.packageName,
      title: vuln.title,
      published: vuln.publicationTime,
      credit: vuln.credit,
      id: vuln.id,
      ids: vuln.identifiers,
      vulnerable: vuln.semver.vulnerable,
      unaffected: vuln.semver.unaffected,
      description: vuln.description,
      apps: new util.SerializableMap(),
      patches: vuln.patches.reduce((total, patch) => {
        if (semver.satisfies(pkg.version, patch.version)) {
          total = total.concat(patch.urls)
        }
        return total
      }, [])
    }
    formattedVulns.set(vuln.id, formattedVuln)
  }
  return formattedVuln
}

function getVulnerabilityApp (vuln, agent) {
  if (!vuln || !agent || !agent.tags) return

  let app = vuln.apps.get(agent.tags.app)
  if (!app) {
    app = {
      appname: agent.tags.app,
      processes: new util.SerializableMap()
    }
    vuln.apps.set(agent.tags.app, app)
  }
  return app
}

function getAppProcess (app, agent) {
  if (!app || !agent) return

  let process = app.processes.get(agent.id)
  if (!process) {
    process = {
      id: agent.id,
      hostname: agent.info ? agent.info.hostname : undefined,
      pid: agent.info ? agent.info.pid : undefined,
      loaded: []
    }
    app.processes.set(agent.id, process)
  }
  return process
}

// remove bits from the packages info structure we don't need
function trimPackages (pkgs) {
  pkgs = JSON.parse(JSON.stringify(pkgs))
  for (let pkg of pkgs) {
    delete pkg.modules
    delete pkg.dependencies
  }
  return pkgs
}

// Normalize the names of the snyk flag files.
// Used to contain colons, now using hyphens.
function fixFlagFileNames (packages) {
  for (let pkg of packages) {
    if (pkg.flagFiles == null) continue

    for (let i = 0; i < pkg.flagFiles.length; i++) {
      pkg.flagFiles[i] = pkg.flagFiles[i].replace(/:/g, '-')
    }
  }
}
