'use strict'

const util = require('./util')
const Logger = util.getLogger(__dirname, __filename)
const lodash = require('lodash')

exports.create = create

const unknownAppError = "Can't determine app name for webhook."
const WEBHOOKS_CONFIG_SETTINGS_KEY = 'webhooks'

function create (context) {
  return new WebhooksManager(context)
}

/*
 * TODO
 * nsolid-threshold-healthy
 *
 * Done
 * nsolid-agent-online
 * nsolid-agent-offline
 * nsolid-profile-generated
 * nsolid-snapshot-generated
 * nsolid-cpu-threshold-exceeded
 * nsolid-heap-threshold-exceeded
 * nsolid-vulnerability-detected
 * nsolid-process-blocked
 */
class WebhooksManager {
  constructor (context) {
    this._context = context

    this._context.diskAssetManager.on('add', asset => {
      this.sendAssetGenerated(asset)
    })
  }

  sendThresholdExceeded (params, cb) {
    cb = cb || function () {}
    const app = getAppFromAgent(params.agent)
    if (!app) {
      return Logger.debug(new Error(unknownAppError))
    }

    const webhookParams = {
      event: `nsolid-${params.metric}-threshold-exceeded`,
      app: app,
      agent: formatAgentInfo(params.agent),
      threshold: params.threshold,
      assets: params.assets
    }

    this.send(webhookParams, cb)
  }

  // This one can be sent on agent startup, so we're using the cb
  // to unset isNotifying in the vulnerability manager to prevent
  // duplicates.
  sendVulnerabilityDetected (params, cb) {
    cb = cb || function () {}
    const app = getAppFromAgent(params.agent)
    if (!app) {
      const err = new Error(unknownAppError)
      Logger.debug(err)
      return cb(err)
    }

    const webhookParams = {
      event: 'nsolid-vulnerability-detected',
      app: app,
      agent: formatAgentInfo(params.agent),
      vulnerability: params.vulnerability
    }

    this.send(webhookParams, cb)
  }

  sendAssetGenerated (asset, cb) {
    cb = cb || function () {}
    const agent = this._context.agentManager.byId(asset.id)
    if (!agent) {
      Logger.debug(`Can't send webhook for ${asset.id}; agent does not exist.`)
      return
    }

    const app = getAppFromAgent(agent)
    if (!app) {
      return Logger.debug(new Error(unknownAppError))
    }

    const params = {
      event: `nsolid-${asset.type}-generated`,
      agent: formatAgentInfo(agent),
      app: app,
      time: Date.now()
    }
    params[asset.type] = lodash.pick(asset, [
      'asset', 'type', 'size', 'compressed'
    ])

    this.send(params, cb)
  }

  sendAgentOffline (agent, cb) {
    cb = cb || function () {}
    const app = getAppFromAgent(agent)
    if (!app) {
      return Logger.debug(new Error(unknownAppError))
    }

    const params = {
      event: 'nsolid-agent-offline',
      app: app,
      agent: formatAgentInfo(agent)
    }

    this.send(params, cb)
  }

  sendAgentOnline (agent, cb) {
    cb = cb || function () {}
    const app = getAppFromAgent(agent)
    if (!app) {
      return Logger.debug(new Error(unknownAppError))
    }

    const params = {
      event: 'nsolid-agent-online',
      app: app,
      agent: formatAgentInfo(agent)
    }

    this.send(params, cb)
  }

  sendProcessBlocked (params, cb) {
    cb = cb || function () {}
    const app = getAppFromAgent(params.agent)
    if (!app) {
      return Logger.error(new Error(unknownAppError))
    }

    const webhookParams = {
      event: 'nsolid-process-blocked',
      app: app,
      agent: formatAgentInfo(params.agent),
      stacktrace: params.stacktrace
    }
    this.send(webhookParams, cb)
  }

  send (params, cb) {
    // when shutting down, webhooks in context will be null
    if (this._context.webhooks == null) return

    params.time = Date.now()

    Logger.debug('send webhooks', params)

    this._context.webhooks.eventsManager.handleEvent(params, cb)
  }

  getConfig () {
    const settingsManager = this._context.settingsManager
    const settings = settingsManager.getSettings()
    const config = settings[WEBHOOKS_CONFIG_SETTINGS_KEY] || {}
    return config
  }

  setConfig (config) {
    const settingsManager = this._context.settingsManager
    const settings = settingsManager.getSettings()
    settings[WEBHOOKS_CONFIG_SETTINGS_KEY] = lodash.cloneDeep(config)

    settingsManager.setSettings(settings, (err) => {
      if (err) {
        Logger.error(err, 'error setting webhooks config')
      }
    })
  }
}

function formatAgentInfo (agent) {
  // This should not happen, but in case a late (post agent exit) lookup
  // results in an undefined agent, we need to do something here
  if (!agent) {
    return {}
  }

  const info = lodash.clone(agent.info || agent)
  delete info.versions
  return info
}

function getAppFromAgent (agent) {
  // This should not happen, but in case a late (post agent exit) lookup
  // results in an undefined agent, don't crash
  if (!agent) {
    return
  }

  if (agent.info && agent.info.app) {
    return agent.info.app
  } else if (agent.app) {
    return agent.app
  } else if (agent.tags && agent.tags.app) {
    return agent.tags.app
  }
}
