'use strict'

const util = require('../util')
const zmq = util.requireZmq()

const nsolidStorage = require('../../nsolid-storage')

const Logger = util.getLogger(__dirname, __filename)

class ZMQRemoteSocket {
  constructor (params) {
    this.remoteUrl = params.remoteUrl
    this.type = params.type
    this.serverKey = params.publicKey
    this.hwm = params.hwm

    if (!/:\/\//.test(this.remoteUrl)) {
      this.remoteUrl = 'tcp://' + this.remoteUrl
    }

    if (/^(pub|pull)$/.test(this.type)) {
      Logger.error(`Use ZMQBindSocket for ${this.type} sockets (${params.remoteUrl || params.bindUrl})`)
      nsolidStorage.shutdown(1)
    }

    this.makeSocket()
  }

  makeSocket () {
    this.socket = zmq.socket(this.type)

    if (this.type === 'sub') {
      Logger.debug('sub socket Subscribing to wildcard ""')
      // subscribe to everything
      this.socket.subscribe('')
    }

    this.socket.monitor(500, 0)
    this.socket.setsockopt(zmq.ZMQ_SNDHWM, this.hwm)
    this.socket.setsockopt(zmq.ZMQ_IPV6, 1)

    const keypair = zmq.curveKeypair()
    this.socket.curve_publickey = keypair.public
    this.socket.curve_secretkey = keypair.secret
    this.socket.curve_serverkey = this.serverKey

    this.socket.on('connect', () => Logger.info(`${this.type} socket connected to ${this.remoteUrl}`))

    this.socket.connect(this.remoteUrl)
  }

  setOptions (remoteUrl, serverKey) {
    this.serverKey = serverKey
    this.remoteUrl = remoteUrl
    if (this.socket) {
      this.socket.disconnect(this.remoteUrl)
      this.socket.curve_serverkey = this.serverKey
      this.socket.connect(this.remoteUrl)
    } else {
      this.makeSocket()
    }
  }

  close () {
    // Handle idempotently
    if (this.socket) {
      this.socket.removeAllListeners()
      this.socket.unmonitor()
      // XXX(Fishrock): I do not know why this did not work like "bind" sockets,
      // but the below code only works under NSolid's bundled zmq.
      // this.socket.on('close', () => {
      //   this.socket.removeAllListeners()
      //   this.socket.unmonitor()
      //   this.socket = null
      // })
      this.socket.setsockopt(zmq.ZMQ_LINGER, 0)
      this.socket.disconnect(this.remoteUrl)
      this.socket.close()
    }
    this.socket = null
  }
}

module.exports = ZMQRemoteSocket
