'use strict'

const os = require('os')
const v8 = require('v8')
const util = require('../util')

const Logger = util.getLogger(__dirname, __filename)

// 4.5 million total in 64-bit, 2.2 million total in 32-bit.
const MaxCombinedLinkedlistLength = os.arch() === 'x64' ? 4500000 : 2200000

// Leave about 3GB
// XXX(Fishrock123): This should be enough for the VM and OS?
const NonAllocatedSpace = 3000
const OneGB = 1000

class SharedMemoryManager {
  constructor (config) {
    // Intialize these on the object in the constructor, but set via setOptions
    this.cleanupSize = 0
    this.memlimit = 0
    this.setOptions(config)

    this.queues = []
    this.listSize = 0
    this.bufferSize = 0

    if (Logger.getLogLevel() === 'debug') {
      // Only launch in debug "mode", the manager has no concept of sockets
      this.logSizeInterval = config.logSizeInterval * 1000 // Convert to ms
      if (this.logSizeInterval < 0 || Number.isNaN(this.logSizeInterval)) {
        this.logSizeInterval = 0
      }

      // bufferSize interval logger, using captured scope for convinence
      const logBufferSize = () => {
        const bufSizeMb = (this.bufferSize / 1000 / 1000).toFixed(3)
        // Only log if size is "significant".
        if (bufSizeMb >= 0.001) {
          Logger.info(`${bufSizeMb}mb shared`)
        }
      }

      // Start interval logger if configured.
      if (this.logSizeInterval) {
        this.loggingTimer = setInterval(logBufferSize, this.logSizeInterval)
        this.loggingTimer.unref()
      }
    }
  }

  setOptions (config) {
    // Total system mem in MB
    const mem = (os.totalmem() / 1000 / 1000)

    this.memlimit = !config.maxBufferSizeMb || config.maxBufferSizeMb < 0 ? mem - NonAllocatedSpace : config.maxBufferSizeMb

    // Not enough space
    if (this.memlimit > mem - OneGB) this.memlimit = mem - OneGB
    if (this.memlimit < 0) this.memlimit = 0

    Logger.info(`${this.memlimit.toFixed(3)}mb automatically allocated to buffering the data & bulk channels`)

    this.memlimit *= 1000 * 1000

    this.cleanupSize = config.cleanupSizeMb // In Megabytes
    if (this.cleanupSize < 0.001 || Number.isNaN(this.cleanupSize)) {
      this.cleanupSize = 0.001
    }

    if (this.bufferSize > this.memlimit) {
      this.cleanupBuffers()
    }
  }

  // Prioritizes in the order added
  registerBufferQueue (queue) {
    this.queues.push(queue)

    // Causes the buffer-queue to emit events for cleanup rather than
    // do cleanup on itself
    queue.shared = true

    queue.on('append', (appendSize) => {
      this.onAppend(appendSize)
    })

    queue.on('flush', (flushSize) => {
      this.onFlush(flushSize)
    })
  }

  onAppend (appendSize) {
    this.listSize += 1
    this.bufferSize += appendSize

    if (this.listSize > MaxCombinedLinkedlistLength) {
      Logger.warn('Combined linkedlist length maxed out! Cleaning up to free VM memory.')

      const stats = v8.getHeapSpaceStatistics()
      for (let i = 0; i < stats.length; i++) {
        if (stats[i].space_name !== 'old_space' && stats[i].space_name !== 'new_space') continue
        Logger.warn(`${stats[i].space_name} mem stats: ${stats[i].space_available_size} bytes free, ${stats[i].space_size} bytes allocated`)
      }

      const goalCount = 10000
      let progress = 0
      let queueIndex = this.queues.length

      while (progress < goalCount && --queueIndex >= 0) {
        const result = this.queues[queueIndex].cleanupLinkedListLength(goalCount - progress)

        progress += result.count
        this.listSize -= result.count
        this.bufferSize -= result.progressBytes
      }
    }

    if (this.bufferSize > this.memlimit) {
      this.cleanupBuffers()
    }
  }

  cleanupBuffers () {
    const goalSize = this.cleanupSize * 1000 * 1000
    let progress = 0
    let queueIndex = this.queues.length

    while (progress < goalSize && --queueIndex >= 0) {
      const result = this.queues[queueIndex].cleanupLinkedList(goalSize - progress)

      progress += result.progressBytes
      this.listSize -= result.count
      this.bufferSize -= result.progressBytes
    }
  }

  onFlush (flushSize) {
    this.listSize -= 1
    this.bufferSize -= flushSize
  }
}

module.exports = SharedMemoryManager
