'use strict'

const Agent = require('../../agent')
const util = require('../../util')
const Logger = util.getLogger(__dirname, __filename)

var SentRecently = []
const RECENT_WINDOW = 60000
const MAX_RECENT = 3

module.exports = function (context) {
  const name = 'loop_blocked'
  const endName = 'loop_unblocked'
  function run (messageObj) {
    let agent = context.agentManager.byId(messageObj.agentId)

    if (!agent) {
      agent = Agent.create(context, messageObj.agentId)
      context.agentManager.upsert(agent)
    }

    if (!messageObj.body) {
      Logger.warn('Got malformed blocked message.', messageObj, agent)
      return
    }

    // Remove items that are no longer recent
    SentRecently = SentRecently.filter(elem => {
      return elem > Date.now() - RECENT_WINDOW
    })

    var info = agent.info || {}
    if (messageObj.command === 'loop_blocked') {
      var formatted = `Process '${info.app || '[unknown app]'}' (${info.pid || '[unknown pid]'}) on ${info.hostname || '[unknown hostname]'} blocked for ${messageObj.body.blocked_for} ms`

      var count = 0
      for (var frame of messageObj.body.stack) {
        if (frame.is_eval) {
          if (frame.script_name) {
            formatted += `\n    at [eval] (${frame.script_name}:${frame.line_number}:${frame.column})`
          } else {
            formatted += `\n    at [eval]:${frame.line_number}:${frame.column}`
          }
        } else {
          if (frame.function_name) {
            formatted += `\n    at ${frame.function_name} (${frame.script_name}:${frame.line_number}:${frame.column})`
          } else {
            formatted += `\n    at ${frame.script_name}:${frame.line_number}:${frame.column}`
          }
        }

        // Limit notification to 10 frames
        if (++count >= 10) {
          break
        }
      }

      Logger.warn(formatted)

      if (SentRecently.length >= MAX_RECENT) {
        Logger.warn('Too many notifications sent, skipping')
        return
      }

      if (hasHooks(context, agent)) {
        SentRecently.push(Date.now())

        context.webhooksManager.sendProcessBlocked({
          agent: agent,
          stacktrace: formatted
        })
      }
    } else {
      const blockedFor = messageObj.body.blocked_for
      const message = `Process ${info.app || '[unknown app]'} (${info.pid || '[unknown pid]'}) on ${info.hostname || '[unknown hostname]'} became unblocked after ${blockedFor} ms`
      Logger.warn(message)
    }
  }

  function predicate (messageObj) {
    return (messageObj.command === name || messageObj.command === endName)
  }

  return { name, run, predicate }
}

function hasHooks (context, agent) {
  let config = context.settingsManager.getSettings().webhooks
  let app = agent.info ? agent.info.app : null
  if (app && config && config[app] && config[app]['nsolid-process-blocked'] &&
      config[app]['nsolid-process-blocked'].webhook) {
    var hooks = config[app]['nsolid-process-blocked'].webhook.filter(hook => {
      return (hook.action === 'http' || hook.action === 'slack') &&
        hook.enabled && hook.url
    })

    return hooks.length > 0
  }
  return false
}
