'use strict'

const request = require('client-request')
const lodash = require('lodash')
const RetryError = require('../retry-error')

module.exports.run = run
module.exports.type = 'http'

// Allow body and reqParams to be override so we can use this
// for Slack API http requests as well.
function run (params, cb) {
  const task = params.task
  const reqBody = lodash.clone(task.evt)
  delete reqBody.customerId
  delete reqBody.action

  const reqParams = params.reqParams || {
    uri: task.config.url,
    headers: {
      'Content-Type': 'application/json'
    },
    method: 'POST',
    body: reqBody
  }

  request(reqParams, (err, response, body) => {
    if (body) {
      body = body.toString()
      try {
        body = JSON.parse(body)
      } catch (err) {
        // ignore err, pass body as-is
      }
    }

    const report = {
      res: response,
      body: body
    }

    if (err) {
      // No retry
      return cb(new RetryError(err.message, false), report)
    }

    if (response.statusCode >= 500 || response.statusCode === 429) {
      // Need to retry
      const msg = `Webhook response status code ${response.statusCode}`
      return cb(new RetryError(msg, true), report)
    } else if (response.statusCode >= 300) {
      // No retry
      const msg = `Webhook response status code ${response.statusCode}`
      return cb(new RetryError(msg, false), report)
    } else if (response.statusCode >= 200 && response.statusCode < 300) {
      // success
      return cb(null, report)
    }

    const msg = `Unknown response ${response.statusCode}`
    return cb(new RetryError(msg, false), report)
  })
}

class IntervalGenerator {
  constructor (params) {
    this.report = null
  }

  // HTTP action interval is a back-off algorithm detailed below unless
  // the customer's API sends us a Retry-After header in a 429 response.
  getInterval (count) {
    if (this.report && this.report.res.statusCode === 429) {
      const num = getRetryAfterMillis(this.report.res)
      if (num !== null) return num
    }

    // roughly equivalent to:
    // 1. no delay
    // 2. 30 secs
    // 3. 60 secs
    // 4. 90 secs
    // 5. 120 secs
    // Never gets called with count === 1
    return count * 30 * 1000
  }

  saveReport (report) {
    this.report = report
  }
}

module.exports.IntervalGenerator = IntervalGenerator

// for HTTP webhooks, we want to fall back to our normal backoff interval
// if Retry-After is not set.
function getRetryAfterMillis (res) {
  if (!res.headers) return null
  const num = parseInt(res.headers['retry-after'])
  if (isNaN(num)) return null
  return num * 1000
}
