'use strict'

const buildSlackMessage = require('./slack-messages')
const httpAction = require('./http')
const formurlencoded = require('form-urlencoded')

module.exports.run = run
module.exports.type = 'slack'

/*
 * Build a slack message and use the HTTP action to send it.
 *
 * params.task:
 *
 * {
 *   evt: {}, // Event object as documented in design doc
 *   config: {} // Customer configuration for this event
 * }
 */
function run (params, cb) {
  const urlParts = params.task.config.url.split('?')
  const passedQstring = urlParts[1] ? `${urlParts[1]}&` : ''
  let message
  var encodedMessage

  try {
    message = buildSlackMessage(params.task)
  } catch (err) {
    return cb(err)
  }

  if (/hooks.slack.com/.test(urlParts[0])) {
    // See https://api.slack.com/incoming-webhooks
    const wrappedMessage = `{"text":"${message}"}`
    encodedMessage = formurlencoded({ payload: wrappedMessage })
  } else {
    encodedMessage = passedQstring + formurlencoded({ text: message })
  }

  params = {
    reqParams: {
      uri: urlParts[0],
      headers: {
        'Content-Type': 'application/x-www-form-urlencoded'
      },
      method: 'POST',
      body: encodedMessage
    },
    task: params.task
  }

  httpAction.run(params, cb)
}

// Slack message interval is 1.5 seconds unless Slack sends us something
// different in the Retry-After header in a 429 response.
class IntervalGenerator {
  constructor (params) {
    this.task = params.task
    this.report = null
    this.rateLimited = params.rateLimited
    if (!this.rateLimited) throw new Error('no rate limited map in constructor')
  }

  /*
   * An algorithm that async.retry will use to determine how long to wait
   * between retries. If we're not currently being rate-limited, we'll just
   * wait 1.5 seconds.
   *
   * If we're currently being rate-limited, we need to determine how severe
   * the rate-limiting is. A computationally easy way to do this is to
   * observe how many times the current task has been retried and how many
   * similar tasks are currently running. It would be more complicated
   * to determine how many similar tasks are waiting to be run.
   *
   * This algorithm was designed with trial and error and results in a 92.5%
   * success rate when sending 400 messages in 90 seconds.
   */
  getInterval (count) {
    const rlCount = this.rateLimited.get(this.task.config.url) || 1
    let delay = 1000 * count
    if (this.report && this.report.res.statusCode === 429) {
      delay = getRetryAfterMillis(this.report.res) + (rlCount * 10)
    }
    return delay
  }

  saveReport (report) {
    this.report = report
  }
}

module.exports.IntervalGenerator = IntervalGenerator

// For Slack messages, default to 1.5 seconds if Retry-After is not set.
function getRetryAfterMillis (res) {
  if (!res.headers) return 1500
  const num = parseInt(res.headers['retry-after'])
  if (isNaN(num)) return 1500
  return num * 1500
}
