'use strict'

const uuid = require('uuid')
const async = require('async')
const request = require('client-request')
const RetryError = require('./retry-error')
const util = require('./util')
const Logger = util.getLogger(__dirname, __filename)

let manager

exports.get = context => {
  if (manager) {
    return manager
  }

  manager = new ApiKeyManager(context)
  return manager
}

class ApiKeyManager {
  constructor (context) {
    this._context = context
    this.ttl = 1000 * 60 * 30 // TODO make configurable
    this.licenseUrl = context.env === 'prod'
      ? 'https://services.nodesource.io/v2/licenses'
      : `https://${context.env}-services.nodesource.io/v2/licenses`
  }

  validateApiKey (params, cb) {
    if (!params.apiKey) {
      setImmediate(() => {
        cb(new Error('Missing parameter `apiKey`.'))
      })
      return
    }

    this._context.db.selectApiKey(params, (err, res) => {
      if (err) {
        return cb(err)
      }

      if (!res) {
        return cb(new Error('No such API Key.'))
      }

      if (res.expiration < Date.now()) {
        return cb(new Error('API Key is expired.'))
      }

      return cb(null, res)
    })
  }

  validateLicense (params, cb) {
    if (!params.licenseKey) {
      setImmediate(() => {
        cb(new Error('Missing parameter `licenseKey`.'))
      })
      return
    }

    async.retry({
      times: 3,
      interval: 10 * 1000,
      errorFilter: err => err.retry
    }, retryCb => {
      request({
        uri: `${this.licenseUrl}/${params.licenseKey}`,
        method: 'GET'
      }, (err, response, body) => {
        if (body) {
          body = body.toString()
          try {
            body = JSON.parse(body)
          } catch (err) {
            Logger.error('Invalid JSON in response from license service.')
          }
        }
        Logger.debug('license validation', err, typeof response.statusCode, body)
        if (err) {
          // No retry
          return retryCb(new RetryError(err.message, false))
        }

        if (response.statusCode >= 500) {
          // Need to retry
          const msg = `Webhook response status code ${response.statusCode}`
          return retryCb(new RetryError(msg, true))
        } else if (response.statusCode >= 300) {
          // No retry
          const msg = `Webhook response status code ${response.statusCode}`
          return retryCb(new RetryError(msg, false))
        } else if (response.statusCode >= 200 && response.statusCode < 300) {
          // success
          return retryCb(null, body)
        }

        // No retry
        const msg = `Unknown response ${response.statusCode}`
        return retryCb(new RetryError(msg, false))
      })
    }, (err, result) => {
      if (err) {
        Logger.error('Error retrieving license', err)
        return cb(err)
      }

      Logger.debug('GET license', result)
      // TODO validate the license
      return cb(null, result)
    })
  }

  createKey (params, cb) {
    const apiKey = {
      customerId: params.customerId,
      apiKey: uuid.v4(),
      expiration: Date.now() + this.ttl
    }
    this._context.db.insertApiKey(apiKey, err => {
      if (err) {
        Logger.error(err, 'Could not create a new api key.')
        return cb(err)
      }

      cb(null, apiKey)
    })
  }

  validateKey (params, cb) {
    if (!params.apiKey) {
      setImmediate(() => {
        cb(new Error('Missing parameter `apiKey`.'))
      })
      return
    }

    this._context.db.selectApiKey(params, (err, key) => {
      if (err) {
        return cb(err)
      }

      // TODO validate the key
      cb(null, true)
    })
  }
}
