'use strict'

const aws = require('aws-sdk')
const util = require('./util')
const Logger = util.getLogger(__dirname, __filename)
const cache = require('./cache')
const lodash = require('lodash')
const uuid = require('uuid')
const tables = require('../etc/ddb-tables')
const apiKeyDefinition = tables.DYNAMODB_API_KEY_TABLE.Properties
const customerConfigDefinition = tables.DYNAMODB_CUSTOMER_CONFIG_TABLE.Properties
const EventEmitter = require('events')
let conn

exports.get = context => {
  if (!context.hosted) {
    throw new Error('Can not create DynamoDB connection in on-prem mode.')
  }

  if (conn) {
    if (conn.ready === true) {
      setImmediate(() => conn.emit('ready'))
    }
    return conn
  }

  aws.config.update(context.config.db)

  conn = new DatabaseConnection(context)
  return conn
}

class DatabaseConnection extends EventEmitter {
  constructor (context) {
    super()
    this.apiKeyTable = tables.getApiKeyTable(context.env)
    this.customerConfigTable = tables.getCustomerConfigTable(context.env)
    Logger.debug('tables', this.apiKeyTable, this.customerConfigTable)
    this._dynamodb = new aws.DynamoDB(context.config.db)
    this._ddb = new aws.DynamoDB.DocumentClient()
    this.webhooks = new cache.WebhooksCache()
    this.apiKeys = new cache.ApiKeysCache()
    this.ready = false

    createTables({
      dynamodb: this._dynamodb,
      apiKeyDefinition: apiKeyDefinition,
      customerConfigDefinition: customerConfigDefinition
    }, function (err) {
      if (err) {
        Logger.error('Could not create the tables', err)
        process.exit()
      }

      Logger.debug('DynamoDB connection ready.', context.config.db)

      setImmediate(function () {
        this.ready = true
        this.emit('ready')
      }.bind(this))
    }.bind(this))
  }

  /*
   * Create a new API key
   */
  insertApiKey (params, cb) {
    Logger.debug('insertApiKey', this.apiKeyTable, params)

    if (!params.apiKey || !params.customerId || !params.expiration) {
      setImmediate(() => {
        cb(new Error('`customerId`, `apiKey`, and `expiration` are required parameters.'))
      })
      return
    }

    const createParams = lodash.pick(params, ['customerId', 'apiKey', 'expiration'])
    this._ddb.put({
      TableName: this.apiKeyTable,
      Item: createParams
    }, err => {
      Logger.debug('insertApiKey results', err, createParams)
      if (err) {
        return cb(err)
      }

      return cb(null, createParams)
    })
  }

  /*
   * Select and return a single API key object from DDB
   */
  selectApiKey (params, cb) {
    Logger.debug('selectApiKey', this.apiKeyTable, params)

    if (!params.apiKey) {
      setImmediate(() => {
        cb(new Error('`apiKey`is a required parameter.'))
      })
      return
    }

    const cached = this.apiKeys.get(params)
    if (cached) {
      setImmediate(() => cb(null, cached))
      return
    }

    this._ddb.get({
      TableName: this.apiKeyTable,
      Key: {
        apiKey: params.apiKey
      }
    }, (err, res) => {
      Logger.debug('selectApiKey results', err, res)
      if (err) return cb(err)
      if (res && res.Item) {
        this.apiKeys.replace({ apiKey: res.Item })
        return cb(null, res.Item)
      }
      cb(null, null)
    })
  }

  /*
   * Create a new webhook configuration entry
   */
  insertWebhook (params, cb) {
    Logger.debug('insertWebhook', this.customerConfigTable, params)

    if (!params.customerId || !params.event || !params.url) {
      setImmediate(() => {
        cb(new Error('`customerId`, `event`, and `url` are required parameters.'))
      })
      return
    }

    const createParams = lodash.pick(params, ['customerId', 'event', 'url'])
    createParams.webhookId = uuid.v4()
    this._ddb.put({
      TableName: this.customerConfigTable,
      Item: createParams
    }, err => {
      if (err) {
        return cb(err)
      }

      return cb(null, createParams)
    })
  }

  /*
   * List and return all webhooks for a particular customer and event.
   */
  listEventWebhooks (params, cb) {
    Logger.debug('listEventWebhooks', this.customerConfigTable, params)

    if (!params.customerId || !params.event) {
      setImmediate(() => {
        cb(new Error('`customerId` and `event` are required parameters.'))
      })
      return
    }

    const cached = this.webhooks.getByEvent(params)
    if (cached && cached.length) {
      setImmediate(() => cb(null, cached))
      return
    }

    this._ddb.query({
      TableName: this.customerConfigTable,
      IndexName: 'customerEvent',
      ConsistentRead: true,
      KeyConditionExpression: 'customerId = :customerId and event = :event',
      ExpressionAttributeValues: {
        ':customerId': params.customerId,
        ':event': params.event
      }
    }, (err, res) => {
      if (err) return cb(err)
      if (res && res.Items) {
        this.webhooks.replace({
          customerId: params.customerId,
          event: params.event,
          webhooks: res.Items
        })
        return cb(null, res.Items)
      }
      cb(null, [])
    })
  }

  /*
   * List and return all webhooks for a particular customer.
   */
  listWebhooks (params, cb) {
    Logger.debug('listWebhooks', this.customerConfigTable, params)

    if (!params.customerId) {
      setImmediate(() => {
        cb(new Error('`customerId` is a required parameter.'))
      })
      return
    }

    const cached = this.webhooks.get(params)
    if (cached && cached.length) {
      setImmediate(() => cb(null, cached))
      return
    }

    this._ddb.query({
      TableName: this.customerConfigTable,
      IndexName: 'customerEvent',
      ConsistentRead: true,
      KeyConditionExpression: 'customerId = :customerId',
      ExpressionAttributeValues: {
        ':customerId': params.customerId
      }
    }, (err, res) => {
      if (err) return cb(err)
      if (res && res.Items) return cb(null, res.Items)
      cb(null, [])
    })
  }

  /*
   * Select and return a single webhook config object by webhookId from DDB
   */
  selectWebhook (params, cb) {
    Logger.debug('selectWebhook', this.customerConfigTable, params)

    if (!params.customerId || !params.webhookId) {
      setImmediate(() => {
        cb(new Error('`customerId` and `webhookId` are required parameters.'))
      })
      return
    }

    const cached = this.webhooks.getById(params)
    if (cached) {
      setImmediate(() => cb(null, cached))
      return
    } else {
      // Hmmm, cache corruption? Delete the entire customer's cache.
      delete this.webhooks[params.customerId]
    }

    this._ddb.get({
      TableName: this.customerConfigTable,
      Key: lodash.pick(params, ['customerId', 'webhookId'])
    }, (err, res) => {
      if (err) return cb(err)
      if (res && res.Item) return cb(null, res.Item)
      cb(null, null)
    })
  }

  /*
   * Delete a single webhook config object by webhookId from DDB
   */
  deleteWebhook (params, cb) {
    Logger.debug('deleteWebhook', this.customerConfigTable, params)

    if (!params.customerId || !params.webhookId) {
      setImmediate(() => {
        cb(new Error('`customerId` and `webhookId` are required parameters.'))
      })
      return
    }

    const cached = this.webhooks.getById(params)
    if (cached) {
      this.webhooks.delete(cached)
    } else {
      // Hmmm, cache corruption? Delete the entire customer's cache.
      delete this.webhooks[params.customerId]
    }

    this._ddb.delete({
      TableName: this.customerConfigTable,
      Key: lodash.pick(params, ['customerId', 'webhookId'])
    }, (err, res) => {
      if (err) return cb(err)
      cb(null, null)
    })
  }
}

function createTables (params, cb) {
  Logger.debug('createTables params', params)
  params.dynamodb.createTable(params.apiKeyDefinition, (err, data) => {
    if (err && err.message.indexOf('exist') === -1) {
      Logger.error(JSON.stringify(err, null, 2))
      return cb(err)
    }
    if (!err) {
      Logger.debug(`created ${params.apiKeyDefinition.TableName}`)
    } else {
      Logger.debug(`${params.apiKeyDefinition.TableName} already exists.`)
    }

    params.dynamodb.createTable(params.customerConfigDefinition, (err, data) => {
      if (err && err.message.indexOf('exist') === -1) {
        Logger.error(JSON.stringify(err, null, 2))
        return cb(err)
      }
      if (err) {
        Logger.debug(`created ${params.customerConfigDefinition.TableName}`)
      } else {
        Logger.debug(`${params.customerConfigDefinition.TableName} already exists.`)
      }
      cb()
    })
  })
}
