'use strict'

exports.handleRequest = handleRequest

const lodash = require('lodash')
const marked = require('marked')

const util = require('../util')
const Logger = util.getLogger(__dirname, __filename)

let ApiDocMarkdown = null
let ApiDocHTML = null

function handleRequest (req, res) {
  if (ApiDocMarkdown == null) {
    ApiDocMarkdown = buildApiDoc(req.nsStorage.context)
  }

  if (ApiDocHTML == null) {
    const markedOpts = {
      gfm: true,
      tables: true,
      sanitize: true
    }
    ApiDocHTML = `${marked(ApiDocMarkdown, markedOpts)}\n${HtmlStyles}`
  }

  res.set('Content-Type', 'text/html')

  if (req.query.markdown != null || req.query.md != null) {
    res.send(`<pre>${ApiDocMarkdown}</pre>`)
  } else {
    res.send(ApiDocHTML)
  }
}

function buildApiDoc (context) {
  const httpSchema = context.schemas.http

  const md = []

  md.push(`# The Totally Unofficial N|Solid Storage Server HTTP API\n`)
  md.push(`generated on ${new Date()}\n`)
  md.push(`[markdown source](?md)\n`)

  for (let method of ['GET', 'PUT', 'POST']) {
    for (let uri in httpSchema[method]) {
      addApiDoc(httpSchema, md, method, uri, httpSchema[method][uri])
    }
  }

  return md.join('\n')
}

function addApiDoc (httpSchema, md, method, uri, uriSchema) {
  md.push('----\n')
  md.push(`## \`${method} /api/v1/${uri}\`\n`)

  if (uriSchema.urlParam) {
    md.push(`## \`${method} /api/v1/${uri}/:${uriSchema.urlParam}\`\n`)
  }

  if (uriSchema.desc) md.push(`${uriSchema.desc}\n`)

  md.push('| query string parm | description |')
  md.push('|-------------------|-------------|')

  if (uriSchema.parms.length === 0) md.push('| none | n/a |')

  for (let parm of getParms(uriSchema.parms)) {
    md.push(`| \`${parm.name}\` | ${parm.desc} |`)
  }

  if (uriSchema.request != null) {
    md.push('\n**request body**\n')
    printRequest(httpSchema, md, uriSchema)
  }

  md.push('\n**response body**\n')
  printResponse(httpSchema, md, uriSchema)
}

function printRequest (httpSchema, md, uriSchema) {
  let request = uriSchema.response

  const requestShapes = getResponseShapes(httpSchema, request)

  for (let shapeName in requestShapes) {
    if (shapeName !== '_') {
      md.push(`sub-shape: {\`${shapeName}\`}\n`)
    }

    md.push('| property | type | description |')
    md.push('|----------|------|-------------|')

    const shape = requestShapes[shapeName]
    for (let prop in shape) {
      let propInfo = splitPropertyLine(shape[prop])

      let type = `\`${propInfo.type}\``
      if (!isPrimitive(propInfo.type)) type = `{${type}}`
      if (propInfo.isArray) type = `[ ${type} ]`

      md.push(`| \`${prop}\` | ${type} | ${propInfo.desc} |`)
    }
    md.push('')
  }
}

function printResponse (httpSchema, md, uriSchema) {
  let isStream = false
  let isArray = false
  let response = uriSchema.response

  if (lodash.isArray(response)) {
    isArray = true
    response = response[0]
  }

  let match = response.match(/^stream:(.*)$/)
  if (match) {
    isStream = true
    response = match[1]
  }

  if (isArray) {
    md.push('Returns an array of the following object shapes:')
    md.push('')
  } else if (isStream) {
    md.push('Returns a stream of the following object shapes:')
    md.push('')
  } else {
    md.push('Returns an object with the following shape:')
    md.push('')
  }

  const responseShapes = getResponseShapes(httpSchema, response)

  for (let shapeName in responseShapes) {
    if (shapeName !== '_') {
      md.push(`sub-shape: {\`${shapeName}\`}\n`)
    }

    md.push('| property | type | description |')
    md.push('|----------|------|-------------|')

    const shape = responseShapes[shapeName]
    for (let prop in shape) {
      let propInfo = splitPropertyLine(shape[prop])

      let type = `\`${propInfo.type}\``
      if (!isPrimitive(propInfo.type)) type = `{${type}}`
      if (propInfo.isArray) type = `[ ${type} ]`

      md.push(`| \`${prop}\` | ${type} | ${propInfo.desc} |`)
    }
    md.push('')
  }
}

// Return an array of response shapes for the specified response.
function getResponseShapes (httpSchema, response) {
  const result = {}

  result._ = httpSchema.type[response]
  if (result._ == null) {
    result._ = { unknown: response }
    return result
  }

  let queue = getResponseSubShapes(result._)

  while (queue.length > 0) {
    const type = queue.shift()
    result[type] = httpSchema.type[type]
    queue = queue.concat(getResponseSubShapes(result[type]))
  }

  return result
}

// Split a property description line:
//   type blah blah description
//   [type] blah blah description
function splitPropertyLine (line) {
  line = line.trim()

  let match = line.match(/^(\S+)(\s(.*))?/)
  if (match == null) {
    Logger.error(`unable to match property line '${line}'`)
    return null
  }

  let type = match[1]
  let desc = match[3] || ''

  match = type.match(/^\[(.*)\]/)
  if (match == null) {
    return { type: type, isArray: false, desc: desc }
  } else {
    return { type: match[1], isArray: true, desc: desc }
  }
}

// Get subtypes of a response
function getResponseSubShapes (responseShape) {
  const result = []

  for (let key in responseShape) {
    let val = splitPropertyLine(responseShape[key])

    if (!isPrimitive(val.type)) {
      result.push(val.type)
    }
  }

  return result
}

const PrimitiveShapes = 'int,float,string,boolean,object'.split(/,/g)

// Return whether type is primitive.
function isPrimitive (type) {
  return PrimitiveShapes.indexOf(type) !== -1
}

// Get the parameters for a specified parameter group
function getParms (parmGroups) {
  const result = []

  for (let parm of parmGroups) {
    let required = false

    if (parm[0] === '!') {
      required = true
      parm = parm.substr(1)
    }

    const reqString = required ? ' **(required)**' : ''

    switch (parm) {
      case 'id':
        result.push({name: 'id', desc: `agent id${reqString}`})
        break
      case 'app':
        result.push({name: 'app', desc: `application name${reqString}`})
        break
      case 'fields':
        result.push({name: 'field', desc: 'output field (may be specified multiple times)'})
        break
      case 'startEnd':
        result.push({name: 'start', desc: 'bound entries starting at specified time'})
        result.push({name: 'end', desc: 'bound entries ending at specified time'})
        break
      case 'interval':
        result.push({name: 'interval', desc: 'display entries by specified interval'})
        break
      case 'agg':
        result.push({name: 'agg', desc: 'aggregation function'})
        break
      case 'duration':
        result.push({name: 'duration', desc: 'duration in seconds'})
        break
      case 'key':
        result.push({name: 'key', desc: 'settings key'})
        break
      case 'val':
        result.push({name: 'val', desc: 'settings value'})
        break
      case 'name':
        result.push({name: 'name', desc: `custom command name${reqString}`})
        break
      case 'data':
        result.push({name: 'data', desc: 'custom command value - any JSONable value'})
        break
      case 'agentScope':
        result.push({name: 'id', desc: 'agent id'})
        result.push({name: 'app', desc: 'application name'})
        result.push({name: 'hostname', desc: 'hostname'})
        result.push({name: 'tag', desc: 'application tag (may be specified multiple times)'})
        break
      case 'asset':
        result.push({name: 'asset', desc: 'asset id'})
        break
      case 'type':
        result.push({name: 'type', desc: 'asset type: snapshot, profile, etc'})
        break
      default:
        throw new Error(`unknown parm ${parm}`)
    }
  }

  return result
}

const HtmlStyles = `
<style>
td, th {
  text-align: left;
  padding: 0.1em 0.5em;
}
</style>
`
