'use strict'

const util = require('./util')
const Logger = util.getLogger(__dirname, __filename)
const EventEmitter = require('events')
const lodash = require('lodash')

const WEBHOOKS_KEY = 'webhooks'
const THRESHOLD_KEY = 'thresholds'
const SECURITY_KEY = 'security'
let conn

// TODO once we get rid of Dynamo creating tables on startup, we can
// export the instance instead of a getter function.
exports.get = context => {
  if (conn) {
    if (conn.ready === true) {
      setImmediate(() => conn.emit('ready'))
    }
    return conn
  }

  conn = new JsonConfig(context)
  return conn
}

class JsonConfig extends EventEmitter {
  constructor (context) {
    super()
    this.ready = true
    this.settingsManager = context.settingsManager
    this.cachedSettings = this.settingsManager.getSettings()
    this.settingsManager.on('settings-changed', () => {
      this.cachedSettings = this.settingsManager.getSettings()
    })
    setImmediate(() => this.emit('ready'))
  }

  /*
   * List and return all webhooks for a particular event.
   * This method is used when sending a webhook, so it's return
   * value must match the same method in db.js.
   */
  listEventWebhooks (params, cb) {
    Logger.debug('listEventWebhooks', params)

    if (!params.event || !params.app) {
      setImmediate(() => {
        cb(new Error('`event` and `app` are required parameters.'))
      })
      return
    }

    const settings = this.cachedSettings
    if (params.event.indexOf('threshold') > -1) {
      return setImmediate(() => {
        listThresholdEventWebhooks(params, settings, cb)
      })
    } else if (params.event.indexOf('vulnerability') > -1) {
      return setImmediate(() => {
        listSecurityEventWebhooks(params, settings, cb)
      })
    } else {
      return setImmediate(() => {
        listNonThresholdEventWebhooks(params, settings, cb)
      })
    }
  }
}

/*
 * Filter non-threshold or security related webhooks config from the settings
 */
function listNonThresholdEventWebhooks (params, settings, cb) {
  if (!settings[WEBHOOKS_KEY]) return cb(null, [])

  const appConfig = settings[WEBHOOKS_KEY][params.app]
  if (!appConfig || !appConfig[params.event]) return cb(null, [])

  const list = parseConfigObj(appConfig[params.event])
  cb(null, list)
}

/*
 * Filter threshold related webhooks config from the settings
 * nsolid-cpu-threshold-exceeded
 * nsolid-heap-threshold-exceeded
 */
function listThresholdEventWebhooks (params, settings, cb) {
  const thresholdType = params.event.indexOf('cpu') > -1 ? 'cpu' : 'heap'

  if (!settings[THRESHOLD_KEY]) return cb(null, [])

  const appConfig = settings[THRESHOLD_KEY][params.app]
  if (!appConfig || !appConfig[thresholdType]) return cb(null, [])

  const list = parseConfigObj(appConfig[thresholdType].actions)
  cb(null, list)
}

/*
 * Filter security related webhooks config from the settings
 * nsolid-vulnerability-detected
 */
function listSecurityEventWebhooks (params, settings, cb) {
  const event = settings[SECURITY_KEY]
  if (!event || !event.actions) return cb(null, [])

  const list = parseConfigObj(event.actions)
  cb(null, list)
}

/*
 * Turn an object with lists of webhook actions into an array
 */
function parseConfigObj (obj) {
  if (!obj || !obj.webhook) return []

  let configs = Array.isArray(obj.webhook) ? obj.webhook : [obj.webhook]
  configs = configs
    .filter(c => c.enabled)
    .map(c => lodash.clone(c))

  return configs
}
