'use strict'

const async = require('async')
const createUuid = require('uuid').v4
const lodash = require('lodash')
const util = require('./util')
const Logger = util.getLogger(__dirname, __filename)
const actions = require('./actions')

let queue
const CONCURRENCY = 30
const SATURATION = 5000 // level at which to drop old tasks
const RATE_LIMIT_SATURATION = 2000 // level at which to drop rate-limited tasks
const rateLimited = new Map()

exports.get = context => {
  if (queue) {
    return queue
  }

  queue = new TaskQueue(context)
  return queue
}

class TaskQueue {
  constructor (context) {
    this._queue = async.queue(this.worker.bind(this), CONCURRENCY)
    this._context = context
    this._queue.drain = () => {
      rateLimited.clear()
    }
  }

  /*
   * Return null if no item was added. If an item was added, return the
   * number of items currently queued and in process.
   *
   * cb is an error handler for this queued task.
   */
  push (evt, config, cb) {
    if (!evt || !evt.event) {
      Logger.warn('no event', evt)
      return cb(new Error('Invalid event.'))
    }

    evt = lodash.defaults({
      eventId: createUuid()
    }, evt)
    Logger.debug('evt', evt)

    if (!config || !config.url) {
      Logger.warn('no config', config)
      return cb(new Error('Invalid config.'))
    }

    this._queue.push({
      evt: evt,
      config: config
    }, cb)
    Logger.debug(`Adding task to queue. Size is ${this.getQueueSize()}`, evt, config)
  }

  getQueueSize () {
    return this._queue.length() + this._queue.running()
  }

  isRateLimitSaturated () {
    return this.getQueueSize() > RATE_LIMIT_SATURATION
  }

  isSaturated () {
    return this.getQueueSize() > SATURATION
  }

  /*
   * By the time a task has been queued, it should have already been validated
   * and matched against customer configuration, which is attached to the task
   * object itself.
   *
   * task:
   *
   * {
   *   evt: {}, // Event object as documented in design doc
   *   config: {} // Customer configuration for this event
   * }
   */
  worker (task, workerCb) {
    const handler = actions.getHandler(task)
    const generator = new handler.IntervalGenerator({
      task: task,
      rateLimited: rateLimited
    })
    let decrement
    let retryOpts = {
      times: 5,
      errorFilter: isRetryable,
      interval: count => generator.getInterval(count)
    }

    // Only increment the map of rate-limited HTTP endpoints the first iteration
    const increment = lodash.once(task => {
      const rlCount = rateLimited.get(task.config.url)
      if (rlCount) {
        rateLimited.set(task.config.url, rlCount + 1)
      } else {
        rateLimited.set(task.config.url, 1)
      }

      decrement = function () {
        let rlCount = rateLimited.get(task.config.url)
        if (rlCount) rlCount--
        if (rlCount < 1) {
          rateLimited.delete(task.config.url)
        } else {
          rateLimited.set(task.config.url, rlCount)
        }
      }
    })

    function runHandler (task, retryCb) {
      handler.run({ task: task }, (err, report) => {
        if (report) {
          if (isRetryable(err)) {
            generator.saveReport(report)
          }

          // this is a once-ed function
          if (report.res.statusCode === 429) {
            increment(task)
          }
        }

        retryCb(err, report)
      })
    }

    Logger.debug(`starting ${handler.type} worker`, task)
    async.retry(retryOpts, retryCb => {
      // Drop the oldest tasks, whether retrying or first time through,
      // if we've hit ${SATURATION} number of tasks queued.
      if (this.isSaturated()) {
        return retryCb(new Error('Dropping task due to queue saturation.'))
      }

      const rlCount = rateLimited.get(task.config.url)
      if (rlCount) {
        if (this.isRateLimitSaturated()) {
          // If this task's HTTP endpoint is currently being rate-limited
          // bc of other identical tasks & we've hit ${RATE_LIMIT_SATURATION}
          // tasks in the queue, drop tasks that would likely be rate-limited
          return workerCb(new Error('Dropping rate-limited task due to queue saturation.'))
        } else {
          // Identical tasks are being rate-limited, but there is a possibility
          // we can recover and continue processing tasks.
          setTimeout(() => {
            runHandler(task, retryCb)
          }, 1500 + (rlCount * 100))
        }
      } else {
        runHandler(task, retryCb)
      }
    }, (err, report) => {
      if (decrement) decrement()
      if (err) {
        Logger.info('Webhook failure', err, task)
      }

      // Passing an error here is currently a noop. We can change this by
      // setting the value of this._queue.error to an error handler.
      workerCb(err, report)
    })
  }
}

function isRetryable (err) {
  return !!(err && err.retry)
}
