#!/usr/bin/env node

'use strict'

exports.main = main
exports.shutdown = shutdown
exports.getConfig = getConfig

const HttpProxyAgent = require('http-proxy-agent')
const HttpsProxyAgent = require('https-proxy-agent')

const pkg = require('./package.json')
const util = require('./lib/util')
const Config = require('./lib/config')

let httpServer

const Logger = util.getLogger(__filename)

const Context = {}

function main (upstreamContext, cb) {
  if (typeof upstreamContext === 'object') {
    Logger.debug('Starting webhooks in on-prem mode')
    Context.settingsManager = upstreamContext.settingsManager
    Context.hosted = false
  } else {
    process.title = 'nsolid-webhooks'
    Logger.debug('Starting webhooks in hosted mode')
    cb = upstreamContext
    upstreamContext = null
    Context.hosted = true

    process.on('SIGINT', shutdownHandler('SIGINT', 0))
    process.on('SIGTERM', shutdownHandler('SIGTERM', 0))
    process.on('uncaughtException', shutdownHandler('uncaughtException', 1))
  }

  Context.version = pkg.version
  Context.env = !process.env.NS_ENV ? 'dev' : process.env.NS_ENV
  Context.mainCB = cb

  if (!Context.config) {
    Context.config = getConfig()
  }

  // wait till help, etc have had a chance to run before adding exit handler
  process.on('exit', onExit)

  logIfHosted(Logger.info, `Env: ${process.env.NS_ENV}`)

  Logger.debug('config:', Context.config)

  logIfHosted(Logger.info, `N|Solid Webhooks Service starting up with PID ${process.pid}`)

  const eventsMan = require('./lib/webhook-events')
  const taskQueueMan = require('./lib/task-queue')

  if (Context.hosted) {
    const ddb = require('./lib/db')
    const apiKeyMan = require('./lib/apikeys')
    Context.db = ddb.get(Context)
    Context.apiKeys = apiKeyMan.get(Context)

    setupHttpAgent(Context)
    httpServer = require('apimule')

    const schema = util.readSchema('http')
    const filepath = `${__dirname}/lib/http-api`
    httpServer.launch(Context, schema, filepath, Logger, (err, proc) => {
      if (Context.db.ready) {
        return httpServerLaunched(err, proc, Context)
      }

      Context.db.once('ready', () => httpServerLaunched(err, proc, Context))
    })
  } else {
    const jsonConfig = require('./lib/json-config')
    Context.db = jsonConfig.get(Context)
    setImmediate(() => httpServerLaunched(null, null, Context))
  }

  Context.queue = taskQueueMan.get(Context)
  Context.eventsManager = eventsMan.get(Context)

  return Context
}

function shutdownHandler (name, code) {
  return (err) => {
    if (Context.hosted) {
      // shutdown doesn't do anything in on-prem mode.
      Logger.error(`stopping process; reason: ${name}`)
      if (err && err instanceof Error) Logger.error(err)
    }

    shutdown(code)
  }
}

function shutdown (code) {
  if (Context.hosted && Context.httpServer) {
    Context.httpServer.shutdown()
    Context.httpServer = null
  }
}

function onExit (code) {
  logIfHosted(Logger.info, `main process exiting with code ${code}`)
}

function httpServerLaunched (err, server, context) {
  if (err) {
    Logger.error(err, 'error launching http server')
    if (Context.mainCB) return Context.mainCB(err)
    shutdown(1)
    return
  }

  if (server) {
    logIfHosted(Logger.info, `launched http server on ${server.url}`)
    context.httpServer = server
    if (context.schemas == null) context.schemas = {}
    if (context.schemas.http == null) context.schemas.http = server.schema
  }

  allLaunched()
}

// Run some code once all subsystems have started
function allLaunched () {
  logIfHosted(Logger.info, 'all subsystems launched!')

  if (Context.mainCB) return Context.mainCB(null, Context)
}

// Set the global http client agent, with HTTP_PROXY override.
function setupHttpAgent (context) {
  const config = context.config

  const httpProxy = config.web.httpProxy
  const httpsProxy = config.web.httpsProxy

  if (httpProxy != null) {
    logIfHosted(Logger.info, `using HTTP proxy: ${httpProxy}`)
    context.httpAgent = new HttpProxyAgent(httpProxy)
  }

  if (httpsProxy != null) {
    logIfHosted(Logger.info, `using HTTPS proxy: ${httpsProxy}`)
    context.httpsAgent = new HttpsProxyAgent(httpsProxy)
  }
}

// get the config, handling -h and -v inline
function getConfig () {
  let conf
  try {
    conf = Config.getConfig()
  } catch (e) {
    Logger.error(e, 'Error attempting to load configuration.')
    process.exit(1)
  }

  // if returned configuration is null, a logged error already printed
  if (conf == null) {
    Logger.error('Failed to load configuration, stopping.')
    process.exit(1)
  }

  if (conf.do_help) {
    util.printHelp()
    process.exit(0)
  }

  if (conf.do_version) {
    console.log(pkg.version)
    process.exit(0)
  }

  if (conf.do_configPrint) {
    delete conf.do_configPrint
    console.log(JSON.stringify(conf, null, 4))
    process.exit(0)
  }

  return conf
}

if (require.main === module) {
  main()
}

function logIfHosted () {
  const args = Array.from(arguments)
  const logger = args.shift()
  if (Context.hosted) {
    logger.apply(null, args)
  }
}
