'use strict'

const test = require('tape')
const testName = 'lib/cache.js'
const caches = require('../lib/cache')
const uuid = require('uuid')

const exampleWebhook = {
  event: uuid.v4(),
  customerId: uuid.v4(),
  webhookId: uuid.v4(),
  url: uuid.v4()
}

test(`${testName} - webhooks cache replace method`, t => {
  const cache = new caches.WebhooksCache()
  t.equals(Object.keys(cache.webhooks).length, 0, 'No customer data cached')

  cache.replace({
    customerId: exampleWebhook.customerId,
    event: exampleWebhook.event,
    webhooks: [exampleWebhook]
  })
  t.equals(Object.keys(cache.webhooks).length, 1, 'One customer in cache')
  const custCache = cache.webhooks[exampleWebhook.customerId]
  t.ok(custCache, 'Keyed by customerId')
  t.equals(typeof custCache, 'object', 'cache is an object')
  t.equals(typeof custCache[exampleWebhook.event], 'object', 'cache contains an event list')
  t.ok(Array.isArray(custCache[exampleWebhook.event]), 'event cache isan array')
  t.equals(custCache[exampleWebhook.event].length, 1, 'event cache has one webhook cached.')
  t.deepEquals(custCache[exampleWebhook.event][0], exampleWebhook, 'the webhook is cached')
  t.deepEquals(custCache.ids[exampleWebhook.webhookId], exampleWebhook, 'the webhook is cached by id, too')
  t.equals(custCache[exampleWebhook.event][0] === custCache.ids[exampleWebhook.webhookId], true, 'webhook is not duplicated in memory')
  t.end()
})

test(`${testName} - webhooks cache delete method`, t => {
  const cache = new caches.WebhooksCache()
  t.equals(Object.keys(cache.webhooks).length, 0, 'No customer data cached')

  cache.replace({
    customerId: exampleWebhook.customerId,
    event: exampleWebhook.event,
    webhooks: [exampleWebhook]
  })

  cache.delete({
    customerId: exampleWebhook.customerId,
    event: exampleWebhook.event
  })
  t.equals(Object.keys(cache.webhooks).length, 1, 'One customer in cache')
  const custCache = cache.webhooks[exampleWebhook.customerId]
  t.ok(custCache, 'Keyed by customerId')
  t.equals(typeof custCache, 'object', 'cache is an object')
  t.notOk(custCache[exampleWebhook.event], 'object', 'cache does not contain an event list')
  t.notOk(custCache.ids[exampleWebhook.webhookId], exampleWebhook, 'the webhook was removed from id cache, too')
  t.end()
})

test(`${testName} - webhooks cache getByEvent method when exists`, t => {
  const cache = new caches.WebhooksCache()
  t.equals(Object.keys(cache.webhooks).length, 0, 'No customer data cached')

  cache.replace({
    customerId: exampleWebhook.customerId,
    event: exampleWebhook.event,
    webhooks: [exampleWebhook]
  })

  const webhooks = cache.getByEvent({
    customerId: exampleWebhook.customerId,
    event: exampleWebhook.event
  })
  t.ok(Array.isArray(webhooks), 'returns an array')
  t.equals(webhooks.length, 1, 'One item in the array.')
  t.deepEquals(webhooks[0], exampleWebhook, 'Array contains the webhook.')
  t.end()
})

test(`${testName} - webhooks cache getByEvent method when customer doesn't exist`, t => {
  const cache = new caches.WebhooksCache()
  t.equals(Object.keys(cache.webhooks).length, 0, 'No customer data cached')

  cache.replace({
    customerId: exampleWebhook.customerId,
    event: exampleWebhook.event,
    webhooks: [exampleWebhook]
  })

  const webhooks = cache.getByEvent({
    customerId: uuid.v4(),
    event: exampleWebhook.event
  })
  t.ok(Array.isArray(webhooks), 'returns an array')
  t.equals(webhooks.length, 0, 'Empty array.')
  t.end()
})

test(`${testName} - webhooks cache getByEvent method when event doesn't exist`, t => {
  const cache = new caches.WebhooksCache()
  t.equals(Object.keys(cache.webhooks).length, 0, 'No customer data cached')

  cache.replace({
    customerId: exampleWebhook.customerId,
    event: exampleWebhook.event,
    webhooks: [exampleWebhook]
  })

  const webhooks = cache.getByEvent({
    customerId: exampleWebhook.customerId,
    event: uuid.v4()
  })
  t.ok(Array.isArray(webhooks), 'returns an array')
  t.equals(webhooks.length, 0, 'Empty array.')
  t.end()
})

test(`${testName} - webhooks cache get method when customer exists`, t => {
  const cache = new caches.WebhooksCache()
  t.equals(Object.keys(cache.webhooks).length, 0, 'No customer data cached')

  cache.replace({
    customerId: exampleWebhook.customerId,
    event: exampleWebhook.event,
    webhooks: [exampleWebhook]
  })

  const webhooks = cache.get({
    customerId: exampleWebhook.customerId
  })
  t.ok(Array.isArray(webhooks), 'returns an array')
  t.equals(webhooks[0], exampleWebhook, 'Contains the webhook.')
  t.end()
})

test(`${testName} - webhooks cache get method when customer exists multipleevent types`, t => {
  const cache = new caches.WebhooksCache()
  t.equals(Object.keys(cache.webhooks).length, 0, 'No customer data cached')

  cache.replace({
    customerId: exampleWebhook.customerId,
    event: exampleWebhook.event,
    webhooks: [exampleWebhook]
  })
  cache.replace({
    customerId: exampleWebhook.customerId,
    event: uuid.v4(),
    webhooks: [exampleWebhook]
  })

  const webhooks = cache.get({
    customerId: exampleWebhook.customerId
  })
  t.ok(Array.isArray(webhooks), 'returns an array')
  t.equals(webhooks.length, 2, 'Contains the webhooks.')
  t.end()
})

test(`${testName} - webhooks cache get method when customer doesn't exist`, t => {
  const cache = new caches.WebhooksCache()
  t.equals(Object.keys(cache.webhooks).length, 0, 'No customer data cached')

  cache.replace({
    customerId: exampleWebhook.customerId,
    event: exampleWebhook.event,
    webhooks: [exampleWebhook]
  })

  const webhooks = cache.get({
    customerId: uuid.v4()
  })
  t.ok(Array.isArray(webhooks), 'returns an array')
  t.equals(webhooks.length, 0, 'Empty array.')
  t.end()
})

test(`${testName} - webhooks cache getById method when webhook exists`, t => {
  const cache = new caches.WebhooksCache()
  t.equals(Object.keys(cache.webhooks).length, 0, 'No customer data cached')

  cache.replace({
    customerId: exampleWebhook.customerId,
    event: exampleWebhook.event,
    webhooks: [exampleWebhook]
  })

  const webhook = cache.getById({
    customerId: exampleWebhook.customerId,
    webhookId: exampleWebhook.webhookId
  })
  t.equal(typeof webhook, 'object', 'returns an object')
  t.deepEquals(webhook, exampleWebhook, 'Returns the right webhook.')
  t.end()
})

test(`${testName} - webhooks cache getById method when webhook doesn't exist`, t => {
  const cache = new caches.WebhooksCache()
  t.equals(Object.keys(cache.webhooks).length, 0, 'No customer data cached')

  cache.replace({
    customerId: exampleWebhook.customerId,
    event: exampleWebhook.event,
    webhooks: [exampleWebhook]
  })

  const webhook = cache.getById({
    customerId: exampleWebhook.customerId,
    webhookId: uuid.v4()
  })
  t.notOk(webhook, 'Does not return a webhook.')
  t.end()
})
