'use strict'

require('./lib/start-dynamodb')
const test = require('tape')
const uuid = require('uuid')
const async = require('async')
const testName = 'lib/db.js'

const context = {
  hosted: true,
  env: 'dev',
  config: {
    db: {
      endpoint: process.env.NSOLID_WEBHOOKS_DB_ENDPOINT || 'http://localhost:8000'
    }
  }
}

let db

test('db is ready', t => {
  db = require('../lib/db').get(context)
  db.once('ready', () => {
    t.pass('db is ready')
    t.end()
  })
})

test(`${testName} insertApiKey with valid params`, t => {
  const createParams = {
    customerId: uuid.v4(),
    apiKey: uuid.v4(),
    expiration: Date.now + (1000 * 30 * 60)
  }
  db.insertApiKey(createParams, err => {
    t.notOk(err, 'No error is returned')

    db._ddb.get({
      TableName: db.apiKeyTable,
      Key: {
        apiKey: createParams.apiKey
      }
    }, (err, res) => {
      t.notOk(err, 'No error when selecting')
      t.ok(res, 'We get a response')
      t.ok(res.Item, 'Key is selectable from the database')
      t.equals(createParams.apiKey, res.Item.apiKey, 'apiKey correct')
      t.equals(createParams.customerId, res.Item.customerId, 'customerId is correct')
      t.equals(createParams.expiration, res.Item.expiration, 'expiration is correct')
      t.end()
    })
  })
})

test(`${testName} insertApiKey with invalid params`, t => {
  const createParams = {
    customerId: uuid.v4(),
    expiration: Date.now + (1000 * 30 * 60)
  }
  db.insertApiKey(createParams, err => {
    t.ok(err, 'An error is returned')
    t.end()
  })
})

test(`${testName} selectApiKey with valid params`, t => {
  const createParams = {
    customerId: uuid.v4(),
    apiKey: uuid.v4(),
    expiration: Date.now + (1000 * 30 * 60)
  }
  db.insertApiKey(createParams, err => {
    t.notOk(err, 'No error is returned')

    db.selectApiKey({
      apiKey: createParams.apiKey
    }, (err, res) => {
      t.notOk(err, 'No error when selecting')
      t.ok(res, 'We get a response')
      t.equals(createParams.apiKey, res.apiKey, 'apiKey correct')
      t.equals(createParams.customerId, res.customerId, 'customerId is correct')
      t.equals(createParams.expiration, res.expiration, 'expiration is correct')
      t.end()
    })
  })
})

test(`${testName} selectApiKey with invalid params`, t => {
  db.selectApiKey({ }, (err, res) => {
    t.ok(err, 'An error is returned')
    t.end()
  })
})

test(`${testName} insertWebhook with valid params`, t => {
  const createParams = {
    customerId: uuid.v4(),
    event: uuid.v4(),
    url: uuid.v4()
  }
  db.insertWebhook(createParams, (err, res) => {
    t.notOk(err, 'No error is returned')

    db._ddb.get({
      TableName: db.customerConfigTable,
      Key: {
        customerId: createParams.customerId,
        webhookId: res.webhookId
      }
    }, (err, res2) => {
      t.notOk(err, 'No error when selecting')
      t.ok(res2, 'We get a response')
      t.ok(res2.Item, 'Key is selectable from the database')
      t.equals(createParams.customerId, res2.Item.customerId, 'customerId is correct')
      t.equals(createParams.event, res2.Item.event, 'event is correct')
      t.equals(createParams.url, res2.Item.url, 'url is correct')
      t.end()
    })
  })
})

test(`${testName} insertWebhook with invalid params`, t => {
  const createParams = {
    url: uuid.v4()
  }
  db.insertWebhook(createParams, err => {
    t.ok(err, 'An error is returned')
    t.end()
  })
})

test(`${testName} selectWebhook with valid params`, t => {
  const createParams = {
    customerId: uuid.v4(),
    event: uuid.v4(),
    url: uuid.v4()
  }
  db.insertWebhook(createParams, (err, res) => {
    t.notOk(err, 'No error is returned')

    db.selectWebhook({
      customerId: createParams.customerId,
      webhookId: res.webhookId
    }, (err, res) => {
      t.notOk(err, 'No error when selecting')
      t.ok(res, 'We get a response')
      t.equals(createParams.customerId, res.customerId, 'customerId is correct')
      t.equals(createParams.event, res.event, 'event is correct')
      t.equals(createParams.url, res.url, 'url is correct')
      t.end()
    })
  })
})

test(`${testName} selectWebhook with invalid params`, t => {
  db.selectWebhook({ webhookId: uuid.v4() }, (err, res) => {
    t.ok(err, 'An error is returned')
    t.end()
  })
})

test(`${testName} listEventWebhooks with valid params`, t => {
  const createParams = {
    customerId: uuid.v4(),
    event: uuid.v4(),
    url: uuid.v4()
  }
  async.parallel([
    cb => db.insertWebhook(createParams, cb),
    cb => db.insertWebhook(createParams, cb),
    cb => db.insertWebhook(createParams, cb)
  ], err => {
    if (err) {
      return t.end(err)
    }

    db.listEventWebhooks({
      customerId: createParams.customerId,
      event: createParams.event
    }, (err, list) => {
      if (err) {
        t.end(err)
      }

      t.ok(list, 'We get a list of webhooks')
      t.equals(list.length, 3, 'There are three webhooks that match')
      list.forEach(wh => {
        t.equals(wh.customerId, createParams.customerId, 'customerId matches')
        t.equals(wh.event, createParams.event, 'event matches')
      })
      t.end()
    })
  })
})

test(`${testName} listEventWebhooks with invalid params`, t => {
  db.listEventWebhooks({
    customerId: uuid.v4()
  }, (err, list) => {
    t.ok(err, 'We get an error')
    t.notOk(list, 'List is not set.')
    t.end()
  })
})

test(`${testName} listWebhooks with valid params`, t => {
  const customerId = uuid.v4()
  const createParams = [{
    customerId: customerId,
    event: uuid.v4(),
    url: uuid.v4()
  }, {
    customerId: customerId,
    event: uuid.v4(),
    url: uuid.v4()
  }, {
    customerId: customerId,
    event: uuid.v4(),
    url: uuid.v4()
  }]
  async.parallel([
    cb => db.insertWebhook(createParams[0], cb),
    cb => db.insertWebhook(createParams[1], cb),
    cb => db.insertWebhook(createParams[2], cb)
  ], err => {
    if (err) {
      return t.end(err)
    }

    db.listWebhooks({
      customerId: customerId
    }, (err, list) => {
      if (err) {
        t.end(err)
      }

      t.ok(list, 'We get a list of webhooks')
      list.forEach(wh => {
        t.equals(wh.customerId, customerId, 'customerId matches')
        t.ok(wh.event, 'event is set')
        t.ok(wh.url, 'url is set')
      })
      t.end()
    })
  })
})

test(`${testName} deleteWebhooks with valid params`, t => {
  db._ddb.scan({
    TableName: db.customerConfigTable
  }, (err, res) => {
    if (err) return t.end(err)
    if (!res || !res.Items) return t.end(new Error('no items'))

    db.deleteWebhook({
      customerId: res.Items[0].customerId,
      webhookId: res.Items[0].webhookId
    }, err => {
      t.notOk(err)

      db.selectWebhook({
        customerId: res.Items[0].customerId,
        webhookId: res.Items[0].webhookId
      }, (err, res) => {
        t.notOk(err, 'No error when selecting')
        t.notOk(res, 'We get no results')
        t.end()
      })
    })
  })
})

test(`${testName} deleteWebhooks with nonexistent webhook`, t => {
  db.deleteWebhook({
    customerId: uuid.v4(),
    webhookId: uuid.v4()
  }, (err, res) => {
    t.notOk(err)
    t.notOk(res, 'We get no results')
    t.end()
  })
})
