'use strict'

const path = require('path')
const ThisFile = path.join(path.basename(__dirname), path.basename(__filename))
const request = require('client-request')
const startServer = require('../start-server')
const integrationUtils = require('../lib/utils')
const uuid = require('uuid')
const nock = require('nock')

startServer.run(runTests)

function runTests (err, context) {
  const test = require('tape')
  if (err) throw err

  const config = context.config.web
  const apiURL = `${config.proto}://${config.server}/api`

  test(`${ThisFile} create api with no license`, t => {
    const params = integrationUtils.getRequestParams({
      uri: `${apiURL}/v1/api-keys`,
      method: 'post',
      json: true
    })

    request(params, (err, res, body) => {
      t.error(err)

      t.equals(res.statusCode, 400, 'API responds 400')
      t.equals(typeof body, 'object', 'body is an object')
      t.equals(body.error.indexOf('validation') > -1, true,
        'License validation error.')
      t.end()
    })
  })

  // TODO make this unecessary by interfacing w/ accounts service
  test(`${ThisFile} create api with valid params invalid license`, t => {
    const customerId = uuid.v4()
    const licenseKey = uuid.v4()
    const params = integrationUtils.getRequestParams({
      uri: `${apiURL}/v1/api-keys`,
      method: 'post',
      json: true,
      body: { customerId: customerId },
      headers: {
        'x-nsolid-license-key': licenseKey
      }
    })

    nock(context.apiKeys.licenseUrl)
      .post(`/{${licenseKey}}`)
      .reply(404)

    request(params, (err, res, body) => {
      t.error(err)

      t.equals(res.statusCode, 200, 'API responds 200')
      t.equals(typeof body, 'object', 'body is an object')
      t.equals(body.customerId, customerId, 'customerId is correct')
      t.ok(body.apiKey, 'apiKey is set')
      t.equals((body.expiration - Date.now()) > 0, true,
        'expiration is in the future')
      t.equals((body.expiration - Date.now()) < (30 * 60 * 1000), true,
        'expiration is not more than 30 mins in the future')
      t.end()
    })
  })

  test(`${ThisFile} create api with valid params valid license`, t => {
    const customerId = uuid.v4()
    const licenseKey = uuid.v4()
    const params = integrationUtils.getRequestParams({
      uri: `${apiURL}/v1/api-keys`,
      method: 'post',
      json: true,
      body: { customerId: customerId },
      headers: {
        'x-nsolid-license-key': licenseKey
      }
    })

    nock(context.apiKeys.licenseUrl)
      .post(`/{${licenseKey}}`)
      .reply(200, {
        licenseKey: licenseKey,
        state: 'VALID'
      })

    request(params, (err, res, body) => {
      t.error(err)

      t.equals(res.statusCode, 200, 'API responds 200')
      t.equals(typeof body, 'object', 'body is an object')
      t.equals(body.customerId, customerId, 'customerId is correct')
      t.ok(body.apiKey, 'apiKey is set')
      t.equals((body.expiration - Date.now()) > 0, true,
        'expiration is in the future')
      t.equals((body.expiration - Date.now()) < (30 * 60 * 1000), true,
        'expiration is not more than 30 mins in the future')
      t.end()
    })
  })
}
