'use strict'

const path = require('path')
const ThisFile = path.join(path.basename(__dirname), path.basename(__filename))
const request = require('client-request')
const startServer = require('../start-server')
const integrationUtils = require('../lib/utils')
const uuid = require('uuid')
const async = require('async')

startServer.run(runTests)

function runTests (err, context) {
  const test = require('tape')
  if (err) throw err

  const config = context.config.web
  const apiURL = `${config.proto}://${config.server}/api`

  test(`${ThisFile} list api with no API key`, t => {
    const params = integrationUtils.getRequestParams({
      uri: `${apiURL}/v1/webhooks`,
      method: 'get',
      json: true
    })

    request(params, (err, res, body) => {
      t.error(err)

      t.equals(res.statusCode, 400, 'API responds 400')
      t.equals(typeof body, 'object', 'body is an object')
      t.equals(body.error.indexOf('validation') > -1, true, 'API validation error.')
      t.end()
    })
  })

  test(`${ThisFile} list api when no webhooks are configured`, t => {
    const params = integrationUtils.getRequestParams({
      uri: `${apiURL}/v1/webhooks`,
      method: 'get',
      json: true,
      headers: {
        'x-nsolid-api-key': context.testApiKey.apiKey
      }
    })

    request(params, (err, res, body) => {
      t.error(err)

      t.equals(res.statusCode, 200, 'API responds 200')
      t.equals(Array.isArray(body), true, 'Body is an array')
      t.equals(body.length, 0, 'Array is empty.')
      t.end()
    })
  })

  test(`${ThisFile} list api when webhooks are configured`, t => {
    const params = integrationUtils.getRequestParams({
      uri: `${apiURL}/v1/webhooks`,
      method: 'get',
      json: true,
      headers: {
        'x-nsolid-api-key': context.testApiKey.apiKey
      }
    })

    const exampleWebhook = {
      customerId: context.testApiKey.customerId,
      event: uuid.v4(),
      url: 'http://invalid.tld'
    }

    context.db.insertWebhook(exampleWebhook, (err, webhook) => {
      t.error(err)

      request(params, (err, res, body) => {
        t.error(err)

        t.equals(res.statusCode, 200, 'API responds 200')
        t.equals(Array.isArray(body), true, 'Body is an array')
        t.notEquals(body.length, 0, 'Array has some items.')

        const wh = body.find(item => item.event === exampleWebhook.event)
        t.equals(wh.customerId, webhook.customerId, 'CustomerID matches')
        t.equals(wh.event, webhook.event, 'Event matches')
        t.equals(wh.url, webhook.url, 'URL matches')
        t.end()
      })
    })
  })

  test(`${ThisFile} get api with no API key`, t => {
    const fakeId = uuid.v4()
    const params = integrationUtils.getRequestParams({
      uri: `${apiURL}/v1/webhook/${fakeId}`,
      method: 'get',
      json: true
    })

    request(params, (err, res, body) => {
      t.error(err)

      t.equals(res.statusCode, 400, 'API responds 400')
      t.equals(body.error.indexOf('validation') > -1, true, 'API validation error.')
      t.end()
    })
  })

  test(`${ThisFile} get api when webhooks are configured`, t => {
    const w1 = {
      customerId: context.testApiKey.customerId,
      event: uuid.v4(),
      url: 'http://invalid.tld'
    }
    const w2 = {
      customerId: context.testApiKey.customerId,
      event: uuid.v4(),
      url: 'http://invalid.tld'
    }
    const w3 = {
      customerId: context.testApiKey.customerId,
      event: uuid.v4(),
      url: 'http://invalid.tld'
    }

    async.parallel({
      w1: cb => context.db.insertWebhook(w1, cb),
      w2: cb => context.db.insertWebhook(w2, cb),
      w3: cb => context.db.insertWebhook(w3, cb)
    }, (err, results) => {
      t.error(err)
      const webhook = results.w1

      const params = integrationUtils.getRequestParams({
        uri: `${apiURL}/v1/webhook/${webhook.webhookId}`,
        method: 'get',
        json: true,
        headers: {
          'x-nsolid-api-key': context.testApiKey.apiKey
        }
      })

      request(params, (err, res, body) => {
        t.error(err)

        t.equals(res.statusCode, 200, 'API responds 200')
        t.equals(typeof body, 'object', 'Body is an object')

        t.equals(body.webhookId, webhook.webhookId, 'WebhookID matches')
        t.equals(body.customerId, webhook.customerId, 'CustomerID matches')
        t.equals(body.event, webhook.event, 'Event matches')
        t.equals(body.url, webhook.url, 'URL matches')
        t.end()
      })
    })
  })

  test(`${ThisFile} get api when the webhook does not exist`, t => {
    const fakeId = uuid.v4()
    const params = integrationUtils.getRequestParams({
      uri: `${apiURL}/v1/webhook/${fakeId}`,
      method: 'get',
      json: true,
      headers: {
        'x-nsolid-api-key': context.testApiKey.apiKey
      }
    })

    request(params, (err, res, body) => {
      t.error(err)

      t.equals(res.statusCode, 404, 'API responds 404')
      t.equals(body.error, 'Not found.', 'Error message is "not found".')
      t.end()
    })
  })

  test(`${ThisFile} create api with no API key`, t => {
    const webhook = {
      event: uuid.v4(),
      url: 'http://invalid.tld'
    }

    const params = integrationUtils.getRequestParams({
      uri: `${apiURL}/v1/webhooks`,
      method: 'post',
      body: webhook,
      json: true
    })

    request(params, (err, res, body) => {
      t.error(err)

      t.equals(res.statusCode, 400, 'API responds 400')
      t.equals(body.error.indexOf('validation') > -1, true, 'API validation error.')
      t.end()
    })
  })

  test(`${ThisFile} create api with valid params`, t => {
    const webhook = {
      event: uuid.v4(),
      url: 'http://invalid.tld'
    }

    const params = integrationUtils.getRequestParams({
      uri: `${apiURL}/v1/webhooks`,
      method: 'post',
      json: true,
      body: webhook,
      headers: {
        'x-nsolid-api-key': context.testApiKey.apiKey,
        'content-type': 'application/json'
      }
    })

    request(params, (err, res, body) => {
      t.error(err)

      t.equals(res.statusCode, 200, 'API responds 200')
      t.equals(typeof body, 'object', 'Body is an object')

      t.ok(body.webhookId, 'WebhookID exists')
      t.equals(body.customerId, context.testApiKey.customerId, 'CustomerID matches')
      t.equals(body.event, webhook.event, 'Event matches')
      t.equals(body.url, webhook.url, 'URL matches')
      t.end()
    })
  })

  test(`${ThisFile} create api with invalid params`, t => {
    const webhook = {
      url: 'http://invalid.tld'
    }

    const params = integrationUtils.getRequestParams({
      uri: `${apiURL}/v1/webhooks`,
      method: 'post',
      json: true,
      body: webhook,
      headers: {
        'x-nsolid-api-key': context.testApiKey.apiKey,
        'content-type': 'application/json'
      }
    })

    request(params, (err, res, body) => {
      t.error(err)

      t.equals(res.statusCode, 400, 'API responds 400')
      t.end()
    })
  })

  test(`${ThisFile} update api with no api key`, t => {
    const webhookId = uuid.v4()
    const editParams = { url: uuid.v4() }
    const params = integrationUtils.getRequestParams({
      uri: `${apiURL}/v1/webhook/${webhookId}`,
      method: 'post',
      json: true,
      body: editParams
    })

    request(params, (err, res, body) => {
      t.error(err)

      t.equals(res.statusCode, 400, 'API responds 400')
      t.equals(body.error.indexOf('validation') > -1, true, 'API validation error.')
      t.end()
    })
  })

  test(`${ThisFile} update api when webhooks are configured`, t => {
    const exampleWebhook = {
      customerId: context.testApiKey.customerId,
      event: uuid.v4(),
      url: 'http://invalid.tld'
    }

    context.db.insertWebhook(exampleWebhook, (err, webhook) => {
      t.error(err)

      const editParams = { url: uuid.v4() }
      const params = integrationUtils.getRequestParams({
        uri: `${apiURL}/v1/webhook/${webhook.webhookId}`,
        method: 'post',
        json: true,
        body: editParams,
        headers: {
          'x-nsolid-api-key': context.testApiKey.apiKey
        }
      })
      request(params, (err, res, body) => {
        t.error(err)

        t.equals(res.statusCode, 501, 'API responds 501')
        /* t.equals(res.statusCode, 200, 'API responds 200')
        t.equals(typeof body, 'object', 'Body is an object')

        t.equals(body.customerId, webhook.customerId, 'CustomerID matches')
        t.equals(body.webhookId, webhook.webhookId, 'CustomerID matches')
        t.equals(body.event, webhook.event, 'Event matches')
        t.equals(body.url, editParams.url, 'URL matches') */
        t.end()
      })
    })
  })

  test(`${ThisFile} delete api with no api key`, t => {
    const exampleWebhook = {
      customerId: context.testApiKey.customerId,
      event: uuid.v4(),
      url: 'http://invalid.tld'
    }

    context.db.insertWebhook(exampleWebhook, (err, webhook) => {
      t.error(err)

      const params = integrationUtils.getRequestParams({
        uri: `${apiURL}/v1/webhook/${webhook.webhookId}`,
        method: 'delete'
      })

      request(params, (err, res, body) => {
        t.error(err)

        t.equals(res.statusCode, 400, 'API responds 400')

        context.db.selectWebhook({
          customerId: context.testApiKey.customerId,
          webhookId: webhook.webhookId
        }, (err, webhook) => {
          t.error(err)
          t.ok(webhook, 'Webhook is still there.')
          t.end()
        })
      })
    })
  })

  test(`${ThisFile} delete api when webhooks are configured`, t => {
    const exampleWebhook = {
      customerId: context.testApiKey.customerId,
      event: uuid.v4(),
      url: 'http://invalid.tld'
    }

    context.db.insertWebhook(exampleWebhook, (err, webhook) => {
      t.error(err)

      const params = integrationUtils.getRequestParams({
        uri: `${apiURL}/v1/webhook/${webhook.webhookId}`,
        method: 'delete',
        headers: {
          'x-nsolid-api-key': context.testApiKey.apiKey
        }
      })

      request(params, (err, res, body) => {
        t.error(err)

        t.equals(res.statusCode, 200, 'API responds 200')

        context.db.selectWebhook({
          customerId: context.testApiKey.customerId,
          webhookId: webhook.webhookId
        }, (err, webhook) => {
          t.error(err)
          t.equals(webhook, null, 'Webhook is null.')
          t.end()
        })
      })
    })
  })

  test(`${ThisFile} delete api when the webhook does not exist`, t => {
    const webhookId = uuid.v4()

    const params = integrationUtils.getRequestParams({
      uri: `${apiURL}/v1/webhook/${webhookId}`,
      method: 'delete',
      headers: {
        'x-nsolid-api-key': context.testApiKey.apiKey
      }
    })

    request(params, (err, res, body) => {
      t.error(err)

      t.equals(res.statusCode, 200, 'API responds 200')
      t.end()
    })
  })
}
