'use strict'

const test = require('tape')
const testName = 'lib/json-config.js'
const EventEmitter = require('events')

let settings = { webhooks: {} }

class MockSettingsManager extends EventEmitter {
  getSettings () {
    return settings
  }

  setSettings (obj, cb) {
    settings = obj
    this.emit('settings-changed')
    setImmediate(cb)
  }
}

const context = {
  env: 'dev',
  settingsManager: new MockSettingsManager()
}

let db

test('db is ready', t => {
  db = require('../lib/json-config').get(context)
  db.once('ready', () => {
    t.pass('db is ready')
    t.end()
  })
})

test(`${testName} listEventWebhooks with non-threshold event`, t => {
  context.settingsManager.setSettings({
    webhooks: {
      'app-name': {
        'some-event': {
          webhook: [{
            enabled: true,
            url: 'http://example.tld/'
          }, {
            action: 'slack',
            enabled: false,
            url: 'http://example.tld/'
          }]
        }
      }
    }
  }, () => {
    db.listEventWebhooks({
      app: 'app-name',
      event: 'some-event'
    }, (err, list) => {
      if (err) {
        t.end(err)
      }

      t.ok(list, 'We get a list of webhooks')
      t.equals(list.length, 1, 'There is 1 webhooks that match')
      t.end()
    })
  })
})

test(`${testName} listEventWebhooks with threshold event`, t => {
  context.settingsManager.setSettings({
    thresholds: {
      'app-name': {
        cpu: {
          actions: {
            webhook: [{
              enabled: false,
              url: 'http://example.tld/'
            }, {
              action: 'slack',
              enabled: true,
              url: 'http://example.tld/'
            }]
          }
        }
      }
    }
  }, () => {
    db.listEventWebhooks({
      app: 'app-name',
      event: 'nsolid-cpu-threshold-exceeded'
    }, (err, list) => {
      if (err) {
        t.end(err)
      }

      t.ok(list, 'We get a list of webhooks')
      t.equals(list.length, 1, 'There is 1 webhooks that match')
      t.end()
    })
  })
})

test(`${testName} listEventWebhooks with a security event`, t => {
  context.settingsManager.setSettings({
    security: {
      actions: {
        webhook: [{
          enabled: false,
          url: 'http://example.tld/'
        }, {
          action: 'slack',
          enabled: true,
          url: 'http://example.tld/'
        }]
      }
    }
  }, () => {
    db.listEventWebhooks({
      app: 'app-name',
      event: 'nsolid-vulnerability-detected'
    }, (err, list) => {
      if (err) {
        t.end(err)
      }

      t.ok(list, 'We get a list of webhooks')
      t.equals(list.length, 1, 'There is 1 webhooks that match')
      t.end()
    })
  })
})

test(`${testName} listEventWebhooks with invalid params`, t => {
  db.listEventWebhooks({}, (err, list) => {
    t.ok(err, 'We get an error')
    t.notOk(list, 'List is not set.')
    t.end()
  })
})
