'use strict'

const test = require('tape')
const testName = 'Slack Action'
const slackAction = require('../lib/actions/slack')
const nock = require('nock')

const exampleTask = {
  config: {
    url: 'blah?something'
  },
  evt: {
    eventId: 'blah'
  }
}

test(`${testName} - Interval Generator 500 response`, t => {
  const generator = new slackAction.IntervalGenerator({
    task: exampleTask,
    rateLimited: new Map()
  })
  const res = {
    statusCode: 500
  }
  generator.saveReport({
    statusCode: res.statusCode,
    res: res
  })
  const num = generator.getInterval(3)
  t.equals(num, 3000, 'Interval is 1 second * number of iterations.')
  t.end()
})

test(`${testName} - Interval Generator 429 response with Retry-After`, t => {
  const generator = new slackAction.IntervalGenerator({
    task: exampleTask,
    rateLimited: new Map()
  })
  const res = {
    statusCode: 429,
    headers: {
      'retry-after': 5
    }
  }
  generator.saveReport({
    statusCode: res.statusCode,
    res: res
  })
  const num = generator.getInterval(3)
  t.equals(num, (5 * 1500) + (1 * 10), 'Interval matches our algorithm.')
  t.end()
})

test(`${testName} - Interval Generator 429 response with no Retry-After`, t => {
  const generator = new slackAction.IntervalGenerator({
    task: exampleTask,
    rateLimited: new Map()
  })
  const res = {
    statusCode: 429
  }
  generator.saveReport({
    statusCode: res.statusCode,
    res: res
  })
  const num = generator.getInterval(3)
  t.equals(num, (1 * 1500) + (1 * 10), 'Interval matches our algorithm.')
  t.end()
})

test(`${testName} - run`, t => {
  const task = {
    config: {
      url: 'https://example.tld/api/chat.postMessage?token=xxxx-xxxxxxxxxx-xxxxxxxxxxx-xxxxxxxxxxxx-xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx&channel=slack-testing&username=nsolid-bot'
    },
    evt: {
      event: 'agent-online',
      app: 'my-app',
      agent: {
        hostname: 'server.example.tld',
        pid: 12345
      }
    }
  }
  nock('https://example.tld').post('/api/chat.postMessage')
    .reply(200, {})

  slackAction.run({ task }, (err, report) => {
    t.error(err, 'No errors.')
    t.ok(report, 'Report is defined')
    t.ok(report.res, 'Response is defined')
    t.equals(report.res.statusCode, 200, 'statusCode is 200')
    t.end()
  })
})
