'use strict'

const test = require('tape')
const taskQueueLib = require('../lib/task-queue')
const http = require('../lib/actions/http')
const testName = 'lib/task-queue.js'
const nock = require('nock')

class IntervalGenerator {
  getInterval () {
    return 200
  }
  saveReport () {
  }
}

const context = {
  stage: 'staging'
}
const taskQueue = taskQueueLib.get(context)

test(`${testName} prepare`, t => {
  // Make it go faster
  http.IntervalGenerator = IntervalGenerator
  t.end()
})

test(`${testName} - tasks can be queued`, t => {
  const config = {
    url: 'http://google.com',
    event: 'something'
  }
  const evt = {
    event: 'something',
    data: 'some data'
  }
  nock(config.url).post('/')
    .reply(200, {})

  taskQueue.push(evt, config, err => {
    t.notOk(err, 'No error is returned')
    t.end()
  })
  t.equals(taskQueue.getQueueSize() > 0, true, 'push function returns the number of items.')
})

test(`${testName} - invalid tasks not queued`, t => {
  const config = {
    url: 'http://google.com',
    event: 'something'
  }
  const evt = {
    data: 'some data'
  }

  taskQueue.push(evt, config, err => {
    t.ok(err, 'Invalid task was not queued')
    t.end()
  })

  t.equals(taskQueue.getQueueSize(), 0, 'push function returns null.')
})

test(`${testName} - when customer API returns an error`, t => {
  const config = {
    url: 'http://google.com',
    event: 'something'
  }
  const evt = {
    event: 'something',
    data: 'some data'
  }
  nock(config.url).post('/')
    .times(5)
    .reply(500, {})

  taskQueue.push(evt, config, err => {
    t.ok(err, 'The error is returned')
    t.end()
  })
  t.equals(taskQueue.getQueueSize() > 0, true, 'push function returns the number of items.')
})

test(`${testName} - when customer API returns an error and then succeeds`, t => {
  const config = {
    url: 'http://google.com',
    event: 'something'
  }
  const evt = {
    event: 'something',
    data: 'some data'
  }
  nock(config.url).post('/')
    .reply(500, {})
  nock(config.url).post('/')
    .reply(500, {})
  nock(config.url).post('/')
    .reply(200, {})

  taskQueue.push(evt, config, (err, report) => {
    t.notOk(err, 'No error is returned')
    t.equals(report.res.statusCode, 200, 'Status code is 200')
    t.end()
  })
  t.equals(taskQueue.getQueueSize() > 0, true, 'push function returns the number of items.')
})

test(`${testName} - when customer API returns 429`, t => {
  const config = {
    url: 'http://google.com',
    event: 'something'
  }
  const evt = {
    event: 'something',
    data: 'some data'
  }
  nock(config.url).post('/')
    .reply(429, {})
  nock(config.url).post('/')
    .reply(200, {})

  taskQueue.push(evt, config, (err, report) => {
    t.notOk(err, 'No error is returned')
    t.equals(report.res.statusCode, 200, 'Status code is 200')
    t.end()
  })
  t.equals(taskQueue.getQueueSize() > 0, true, 'push function returns the number of items.')
})
