'use strict'

const request = require('client-request')
const async = require('async')
const EventEmitter = require('events')
const uuid = require('uuid')
const quasix = require('quasix-getopt')
const options = quasix.parse()

class WebhooksManager extends EventEmitter {
  constructor () {
    super()

    if (options.aws) {
      this._url = 'https://webhooks-staging.nsolid.nodesource.io/api/v1'
    } else {
      this._url = 'http://localhost:80/api/v1'
    }

    this._apiKey = null
    this._ready = null
    this.license = 'f25a2e30-eb00-48c0-a74b-47c7a23b1bad'

    request({
      uri: `${this._url}/health`,
      method: 'get',
      json: true
    }, (err, res, body) => {
      if (err || res.statusCode !== 200) {
        console.error('Can not access webhooks service', err, res.statusCode)
        process.exit()
      }

      async.retry({
        times: 10,
        interval: 30 * 1000
      }, retryCb => this.refreshApiKey(retryCb), err => {
        if (err) {
          // TODO trigger this again when the license key is changed
          console.error('Can not get a webhooks API key after 10 tries, giving up.')
          return
        }

        this._ready = true
        this.emit('ready')

        setInterval(() => {
          this.refreshApiKey(err => {
            console.log('Could not refresh webhooks API key', err)
          })
        }, 29 * 60 * 1000)
      })
    })
  }

  refreshApiKey (cb) {
    if (!this.license) {
      console.log('Can not refresh webhooks API key: no license.')
      cb(new Error('Can not refresh webhooks API key: no license.'))
      return
    }

    console.log('Attempting to obtain a webhooks API key.', this.license, `${this._url}/api-keys`)
    request({
      uri: `${this._url}/api-keys`,
      method: 'POST',
      headers: {
        'x-nsolid-license-key': this.license
      },
      body: {
        customerId: 'erintest1' // TODO remove this
      }
    }, (err, res, body) => {
      if (err || res.statusCode !== 200) {
        console.log(res.statusCode, err || (body || '').toString())
        return cb(err || new Error(res.statusCode + (body || '').toString()))
      }
      if (body) {
        try {
          body = JSON.parse(body.toString())
        } catch (err) {
          console.error(`Error getting webhooks API key: ${res.statusCode}.`, err, body.toString())
          return
        }
        if (!body.apiKey || body.apiKey === 'null' || body.apiKey === 'undefined') {
          cb(new Error('Got invalid API key: ' + typeof body.apiKey + ' ' + body.apiKey))
          return
        }
        this._apiKey = body.apiKey
        console.log('Successfully refreshed webhooks API key.')
        cb(null, body.apiKey)
        return
      }

      console.error(`Error getting webhooks API key: ${res.statusCode}.`, err, body)
      cb(err)
    })
  }

  send (index) {
    const params = {
      event: 'nsolid-agent-online',
      index: index,
      id: uuid.v4(),
      txTime: Date.now()
    }

    params.apiKey = this._apiKey
    // console.log('send webhooks', params)

    request({
      uri: `${this._url}/events`,
      method: 'POST',
      headers: {
        'x-nsolid-api-key': this._apiKey
      },
      body: params
    }, (err, res, body) => {
      if (body) {
        try {
          body = JSON.parse(body.toString())
        } catch (err) {
          console.error('webhook response is invalid JSON', body.toString())
          return
        }
      }

      if (err || !body || res.statusCode !== 200) {
        // console.error(`Error sending webhooks event ${res.statusCode}.`, err)
        return
      }
      // console.log(`Sent webhook event. ${res.statusCode}`, params, body)
    })

    return params
  }

}

module.exports = WebhooksManager
