#!/usr/bin/env node

'use strict'

exports.main = main
exports.shutdown = shutdown
exports.getConfig = getConfig

const HttpProxyAgent = require('http-proxy-agent')
const HttpsProxyAgent = require('https-proxy-agent')

const pkg = require('./package.json')
const util = require('./lib/util')
const Config = require('./lib/config')

// change 0 to something like 30000 to test long influx startup
const INFLUX_LAUNCH_DELAY_MS = 0

let influxdb
let zmqServer
let httpServer
let agentManager
let commandDispatcher
let assetManager
let diskAssetManager
let vulnManager
let settingsManager
let providerManager
let initThresholdActions
let webhooksManager
let webhooks

const Logger = util.getLogger(__filename)

const Context = {}

function main (cb) {
  process.on('SIGINT', shutdownHandler('SIGINT', 0))
  process.on('SIGTERM', shutdownHandler('SIGTERM', 0))
  process.on('uncaughtException', shutdownHandler('uncaughtException', 1))

  Context.version = pkg.version
  Context.isStaging = (process.env.NSOLID_STAGING != null)
  Context.mainCB = cb

  if (!Context.config) {
    Context.config = getConfig()
  }

  // wait till help, etc have had a chance to run before adding exit handler
  process.on('exit', onExit)

  if (Context.isStaging) Logger.info('using staging')

  Logger.debug('config:', Context.config)

  Logger.info(`N|Solid Storage Service starting up with PID ${process.pid}`)

  setupHttpAgent(Context)

  if (Context.config.relayMode) {
    Logger.info('launching in Relay mode')

    // Lazy-load Relay
    const zmqRelay = require('./lib/zmqrelay')

    zmqRelay.launch(Context, (err, relay) => zmqRelayLaunched(err, relay, Context))

    return Context
  }

  // Lazy load so that we don't load lots of unncessary things for the Relay
  influxdb = require('./lib/influxdb')
  zmqServer = require('./lib/zmqserver')
  httpServer = require('./lib/http-server')
  agentManager = require('./lib/agent-manager')
  commandDispatcher = require('./lib/command-dispatcher')
  assetManager = require('./lib/asset-manager')
  diskAssetManager = require('./lib/disk-asset-manager')
  vulnManager = require('./lib/vuln-manager')
  settingsManager = require('./lib/settings-manager')
  providerManager = require('./lib/provider')
  webhooksManager = require('./lib/webhooks-manager')
  initThresholdActions = require('./lib/threshold-actions').initThresholdActions
  webhooks = require('nsolid-webhooks')

  Context.agentManager = agentManager.create(Context)
  Context.commandDispatcher = commandDispatcher.create(Context)
  Context.assetManager = assetManager.create(Context)
  Context.diskAssetManager = diskAssetManager.create(Context)
  Context.provider = providerManager.create(Context)
  Context.vulnManager = vulnManager.create(Context)
  Context.settingsManager = settingsManager.create(Context)
  Context.webhooksManager = webhooksManager.create(Context)
  Context.webhooks = webhooks.main(Context)

  launchSubSystems()

  return Context
}

// launch the subsytems in parallel, call allLaunched() when all are launched
function launchSubSystems () {
  let countDown = 3

  setTimeout(
    () => {
      influxdb.launch(Context, (err, proc) => {
        influxdbLaunched(err, proc, Context)
        subSystemDone()
      })
    },
    // should be zero, but can be changed at top, to test long influx startups
    INFLUX_LAUNCH_DELAY_MS
  )

  zmqServer.launch(Context, (err, server) => {
    zmqServerLaunched(err, server, Context)
    subSystemDone()
  })

  httpServer.launch(Context, (err, proc) => {
    httpServerLaunched(err, proc, Context)
    subSystemDone()
  })

  function subSystemDone () {
    countDown--
    if (countDown === 0) allLaunched()
  }
}

function shutdownHandler (name, code) {
  return (err) => {
    // ignore errors on stdout, while going down ...
    Logger.ignoreStreamError(true)
    Logger.error(`stopping process; reason: ${name}`)
    if (err && err instanceof Error) Logger.error(err)

    shutdown(code)
  }
}

const INFLUX_SHUTDOWN_MAX_SECONDS = 30
let ShuttingDown = false

function shutdown (code) {
  // ignore errors on stdout, while going down ...
  Logger.ignoreStreamError(true)

  if (ShuttingDown) {
    Logger.error(`shutdown(${code}) called while already shutting down`)
    return
  }

  ShuttingDown = true

  if (Context.zmqRelay) {
    Context.zmqRelay.shutdown()
    Context.zmqRelay = null
  }
  if (Context.zmqServer) {
    Context.zmqServer.shutdown()
    Context.zmqServer = null
  }
  if (Context.httpServer) {
    Context.httpServer.shutdown()
    Context.httpServer = null
  }
  if (Context.diskAssetManager) {
    Context.diskAssetManager.shutdown()
    Context.diskAssetManager = null
  }
  if (webhooks) {
    webhooks.shutdown()
    Context.webhooks = null
  }

  // last stop before having to shutdown influx
  if (!Context.influxdbProcess) {
    Logger.error('InfluxDB did not successfully start; exiting')
    process.exit(code)
  }

  // influxd can take a few seconds to come down, so gate actual exit on it
  Logger.info('waiting for InfluxDB process to shutdown')
  Context.influxdbProcess.kill(() => {
    Logger.info('InfluxDB process shutdown')
    process.exit(code)
  })

  setTimeout(() => {
    Logger.error('forcing shutdown, InfluxDB not finished shutting down')
    process.exit(1)
  }, INFLUX_SHUTDOWN_MAX_SECONDS * 1000).unref()
}

function onExit (code) {
  // ignore errors on stdout, while going down ...
  Logger.ignoreStreamError(true)
  Logger.info(`main process exiting with code ${code}`)
}

function zmqRelayLaunched (err, relay, config) {
  if (err) {
    Logger.error(err, 'error starting ZeroMQ relay')
    if (Context.mainCB) return Context.mainCB(err)
    shutdown(1)
    return
  }

  Logger.info(`Listening to SIGHUP for config reloading on PID ${process.pid}`)

  process.on('SIGHUP', () => {
    setImmediate(() => {
      let conf
      try {
        conf = Config.getConfig()
      } catch (err) {
        Logger.log(`Config reload failed: ${err.message}`)
        return
      }
      if (!conf) {
        Logger.log('Config reload failed, invalid configuration.')
        return
      }

      Context.config = conf
      relay.reloadConfig(Context.config)
    })
  })

  Logger.info('launched ZeroMQ relay')
  Context.zmqRelay = relay

  allLaunched()
}

function influxdbLaunched (err, proc, context) {
  if (err) {
    Logger.error(err, 'error launching InfluxDB')
    if (Context.mainCB) return Context.mainCB(err)
    shutdown(1)
    return
  }

  Logger.info(`launched InfluxDB on ${proc.url}`)
  context.influxdbProcess = proc
  context.influxdbReader = influxdb.createReader(context)
  context.influxdbWriter = influxdb.createWriter(context)
}

function zmqServerLaunched (err, server, config) {
  if (err) {
    Logger.error(err, 'error starting ZeroMQ server')
    if (Context.mainCB) return Context.mainCB(err)
    shutdown(1)
    return
  }

  Logger.debug('launched ZeroMQ server')
  Context.zmqServer = server
}

function httpServerLaunched (err, server, context) {
  if (err) {
    Logger.error(err, 'error launching http server')
    if (Context.mainCB) return Context.mainCB(err)
    shutdown(1)
    return
  }

  Logger.info(`launched http server on ${server.url}`)
  context.httpServer = server
}

// Run some code once all subsystems have started
function allLaunched () {
  if (initThresholdActions) {
    initThresholdActions(Context)
  }

  Logger.info('all subsystems launched!')

  if (Context.mainCB) return Context.mainCB(null, Context)
}

// Set the global http client agent, with HTTP_PROXY override.
function setupHttpAgent (context) {
  const config = context.config

  const httpProxy = config.web.httpProxy
  const httpsProxy = config.web.httpsProxy

  if (httpProxy != null) {
    Logger.info(`using HTTP proxy: ${httpProxy}`)
    context.httpAgent = new HttpProxyAgent(httpProxy)
  }

  if (httpsProxy != null) {
    Logger.info(`using HTTPS proxy: ${httpsProxy}`)
    context.httpsAgent = new HttpsProxyAgent(httpsProxy)
  }
}

// get the config, handling -h and -v inline
function getConfig () {
  let conf
  try {
    conf = Config.getConfig()
  } catch (e) {
    Logger.error(e, 'Error attempting to load configuration.')
    process.exit(1)
  }

  // if returned configuration is null, a logged error already printed
  if (conf == null) {
    Logger.error('Failed to load configuration, stopping.')
    process.exit(1)
  }

  if (conf.do_help) {
    util.printHelp()
    process.exit(0)
  }

  if (conf.do_version) {
    console.log(pkg.version)
    process.exit(0)
  }

  if (conf.do_genkey) {
    genCurveKeypair()
    process.exit(0)
  }

  if (conf.do_configPrint) {
    delete conf.do_configPrint
    console.log(JSON.stringify(conf, null, 4))
    process.exit(0)
  }

  return conf
}

function genCurveKeypair () {
  const zmq = util.requireZmq()
  const pair = zmq.curveKeypair()

  process.stdout.write('#################################\n')
  process.stdout.write('## N|Solid Storage config TOML ##\n')
  process.stdout.write('#################################\n')
  process.stdout.write('[sockets]\n')
  process.stdout.write('\n')
  process.stdout.write(`publicKey = "${pair.public}"\n`)
  process.stdout.write(`privateKey = "${pair.secret}"\n`)
  process.stdout.write('\n')
  process.stdout.write('######################################\n')
  process.stdout.write('## N|Solid Storage environment vars ##\n')
  process.stdout.write('######################################\n')
  process.stdout.write(`NSOLID_STORAGE_SOCKETS_PUBLIC_KEY='${pair.public}'\n`)
  process.stdout.write(`NSOLID_STORAGE_SOCKETS_PRIVATE_KEY='${pair.secret}'\n`)
  process.stdout.write('\n')
  process.stdout.write('#####################################\n')
  process.stdout.write('## N|Solid Runtime environment var ##\n')
  process.stdout.write('#####################################\n')
  process.stdout.write(`NSOLID_PUBKEY='${pair.public}'\n`)
}

if (require.main === module) {
  process.title = 'nsolid-storage'
  main()
}
