#!/usr/bin/env node
'use strict'

process.title = 'nsolid-cli'
const rc = require('rc')
const minimist = require('minimist')
const cli = require('./cli')
const formats = require('./cli_formats')
const packageJSON = require('./package.json')

const bole = require('bole')

const debugEnv = process.env.NODE_DEBUG
var loglevel = 'info'
if (debugEnv != null && debugEnv.indexOf('nsolid') >= 0) {
  loglevel = 'debug'
}

bole.output({
  level: loglevel,
  stream: process.stderr
})

var argv = minimist(process.argv.slice(2), { 'boolean': ['force'] })
var conf = rc('nsolid-cli', {
  remote: 'https://localhost:4000'
}, argv)

if (conf.h || conf.help) {
  const helpFile = require('path').join(__dirname, 'help.txt')
  const helpText = require('fs').readFileSync(helpFile)
  console.error(String(helpText))
  process.exit(0)
}

if (conf.s || conf.stdin) {
  // passing -s or --stdin expects to read from stdin
  conf.body = process.stdin
}

if (conf.v || conf.version) {
  console.log(packageJSON.version)
  process.exit(0)
}

// _secret_ env var (not official, so don't document) in order to easier configure limit for tests
if (typeof conf.limit === 'undefined' && typeof process.env.NSOLID_CLI_LIMIT !== 'undefined') {
  conf.limit = process.env.NSOLID_CLI_LIMIT
  if (conf.limit !== 'auto' && typeof conf.limit !== 'number') {
    conf.limit = parseInt(conf.limit, 10)
  }
  if (conf.limit !== 'auto' && isNaN(conf.limit)) {
    throw new Error(`Supplied invalid NSOLID_CLI_LIMIT '${process.env.NSOLID_CLI_LIMIT}'.`)
  }
}

if (conf._.length === 0) {
  usageError('Expecting a command: ping, snapshot, etc. Use -h for list of commands.')
}

if (conf._.length > 1) {
  usageError('Expecting only a single command')
}

var command = conf._[0]
// 'ls' is an alias for 'list'
if (command === 'ls') {
  command = 'list'
}

var remoteSocket = process.env.NSOLID_STORAGE_ENDPOINT || ('' + conf.remote)
if (remoteSocket.indexOf(':') < 0) {
  remoteSocket = 'localhost:' + remoteSocket
}

try {
  var out = cli(command, remoteSocket, conf)
} catch (err) {
  console.error(err.stack)

  usageError(err.message)
}

out.on('error', function (err) {
  usageError('N|Solid error: ' + err.message)
})

out.on('response', res => {
  const printTo = /^2/.test(res.statusCode) ? process.stdout : process.stderr
  out
    .pipe(formats.fetchAsset(out.fetchAsset, remoteSocket, out))
    .pipe(formats.Consolize(printTo))
    .pipe(printTo)
})

function usageError (message) {
  console.error(message)
  process.exit(1)
}
